# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import re
import typing as t

from ansible.module_utils.common.text.converters import to_text

if t.TYPE_CHECKING:
    from collections.abc import Sequence  # pragma: no cover

_ASCII_PRINTABLE_MATCHER = re.compile(r"^[\x20-\x7e]*$")


def is_ascii_label(domain: str) -> bool:
    """
    Check whether domain name has only ASCII labels.
    """
    return _ASCII_PRINTABLE_MATCHER.match(domain) is not None


class InvalidDomainName(Exception):
    """
    The provided domain name is not valid.
    """


def split_into_labels(domain: str) -> tuple[list[str], t.Literal["", "."]]:
    """
    Split domain name to a list of labels. Start with the top-most label.

    Returns a list of labels and a tail, which is either ``''`` or ``'.'``.
    Raises ``InvalidDomainName`` if the domain name is not valid.
    """
    result = []
    index = len(domain)
    tail: t.Literal["", "."] = ""
    if domain.endswith("."):
        index -= 1
        tail = "."
    if index > 0:
        while index >= 0:
            next_index = domain.rfind(".", 0, index)
            label = domain[next_index + 1 : index]
            if label == "" or label[0] == "-" or label[-1] == "-" or len(label) > 63:
                raise InvalidDomainName(domain)
            result.append(label)
            index = next_index
    return result, tail


def join_labels(labels: Sequence[str], tail: str = "") -> str:
    """
    Combines the result of split_into_labels() back into a domain name.
    """
    return ".".join(reversed(labels)) + tail


def normalize_label(label: str) -> str:
    """
    Normalize a domain label. Returns a lower-case ASCII label.

    If a ulabel is provided, it is converted to an alabel.
    """
    if not is_ascii_label(label):
        # Convert ulabel to alabel
        label = to_text(b"xn--" + to_text(label).encode("punycode"))
    # Always convert to lower-case
    return label.lower()



#
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

"""
The type5_pw plugin code
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

import string

from ansible.errors import AnsibleFilterError
from ansible.module_utils.common.text.converters import to_text


try:
    # old import
    HAS_PASSLIB_OR_CRYPT = True
    from ansible.utils.encrypt import passlib_or_crypt, random_password
except ImportError:
    # To use do_encrypt from 2.20
    from ansible.utils.encrypt import random_password

    HAS_PASSLIB_OR_CRYPT = False

string_types = (str,)


def _raise_error(msg):
    raise AnsibleFilterError(msg)


def type5_pw(password, salt=None):
    if not isinstance(password, string_types):
        _raise_error(
            "type5_pw password input should be a string, but was given a input of %s"
            % (type(password).__name__)
        )

    salt_chars = "".join((to_text(string.ascii_letters), to_text(string.digits), "./"))
    if salt is not None and not isinstance(salt, string_types):
        _raise_error(
            "type5_pw salt input should be a string, but was given a input of %s"
            % (type(salt).__name__)
        )
    elif not salt:
        salt = random_password(length=4, chars=salt_chars)
    elif not set(salt) <= set(salt_chars):
        _raise_error("type5_pw salt used inproper characters, must be one of %s" % (salt_chars))

    if HAS_PASSLIB_OR_CRYPT:
        encrypted_password = passlib_or_crypt(password, "md5_crypt", salt=salt)
    else:
        from ansible.utils.encrypt import do_encrypt

        encrypted_password = do_encrypt(password, "md5_crypt", salt=salt)

    return encrypted_password



# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible_collections.ansible.mysql.plugins.module_utils.mysql import get_server_version
from ansible_collections.ansible.mysql.plugins.module_utils.version import LooseVersion


def uses_replica_terminology(cursor):
    """Checks if REPLICA must be used instead of SLAVE"""
    return LooseVersion(get_server_version(cursor)) >= LooseVersion('8.0.22')



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright: (c) 2023, Gaudenz Steinlin <gaudenz.steinlin@cloudscale.ch>
# Copyright: (c) 2023, Kenneth Joss <kenneth.joss@cloudscale.ch>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: load_balancer_health_monitor
short_description: Manages load balancers on the cloudscale.ch IaaS service
description:
  - Get, create, update, delete health monitors on the cloudscale.ch IaaS service.
notes:
  - Health monitors do not have names. I(uuid)'s are used to reference a health monitors.
author:
  - Gaudenz Steinlin (@gaudenz)
  - Kenneth Joss (@k-304)
version_added: "2.3.0"
options:
  state:
    description:
      - State of the load balancer health monitor.
    choices: [ present, absent ]
    default: present
    type: str
  uuid:
    description:
      - UUID of the load balancer health monitor.
    type: str
  pool:
    description:
      - The pool of the health monitor.
    type: str
  delay_s:
    description:
      - The delay between two successive checks in seconds.
    type: int
  timeout_s:
    description:
      - The maximum time allowed for an individual check in seconds.
    type: int
  up_threshold:
    description:
      - The number of checks that need to be successful before the monitor_status of a pool member changes to "up".
    type: int
  down_threshold:
    description:
      - The number of checks that need to fail before the monitor_status of a pool member changes to "down".
    type: int
  type:
    description:
      - The type of the health monitor.
      - See the [API documentation](https://www.cloudscale.ch/en/api/v1#create-a-health-monitor) for allowed options.
    type: str
  http:
    description:
      - Advanced options for health monitors with type "http" or "https".
    type: dict
    suboptions:
      expected_codes:
        description:
          - The HTTP status codes allowed for a check to be considered successful.
          - See the [API documentation](https://www.cloudscale.ch/en/api/v1#http-attribute-specification) for details.
        type: list
        elements: str
      method:
        description:
          - The HTTP method used for the check.
        type: str
      url_path:
        description:
          - The URL used for the check.
        type: str
      version:
        description:
          - The HTTP version used for the check.
        type: str
      host:
        description:
          - The server name in the HTTP Host header used for the check.
          - Requires version to be set to "1.1".
        type: str
  tags:
    description:
      - Tags assosiated with the load balancer. Set this to C({}) to clear any tags.
    type: dict
extends_documentation_fragment: cloudscale_ch.cloud.api_parameters
'''

EXAMPLES = '''
# Create a simple health monitor for a pool
- name: Create a load balancer pool
  cloudscale_ch.cloud.load_balancer_pool:
    name: 'swimming-pool'
    load_balancer: '3d41b118-f95c-4897-ad74-2260fea783fc'
    algorithm: 'round_robin'
    protocol: 'tcp'
    api_token: xxxxxx
  register: load_balancer_pool

- name: Create a load balancer health monitor (ping)
  cloudscale_ch.cloud.load_balancer_health_monitor:
    pool: '{{ load_balancer_pool.uuid }}'
    type: 'ping'
    api_token: xxxxxx
  register: load_balancer_health_monitor

# Craete a udp health monitor for a udp pool
- name: Create a udp load balancer pool
  cloudscale_ch.cloud.load_balancer_pool:
    name: 'udp-pool'
    load_balancer: '3d41b118-f95c-4897-ad74-2260fea783fc'
    algorithm: 'round_robin'
    protocol: 'udp'
    api_token: xxxxxx
  register: udp_load_balancer_pool

- name: Create a udp load balancer health monitor (udp-connect)
  cloudscale_ch.cloud.load_balancer_health_monitor:
    pool: '{{ udp_load_balancer_pool.uuid }}'
    type: 'udp-connect'
    delay_s: 3
    timeout_s: 2
    api_token: xxxxxx
  register: udp_load_balancer_health_monitor

# Get load balancer health monitor facts by UUID
- name: Get facts of a load balancer health monitor by UUID
  cloudscale_ch.cloud.load_balancer_health_monitor:
    uuid: '{{ load_balancer_health_monitor.uuid }}'
    api_token: xxxxxx

# Update a health monitor
- name: Update HTTP method of a load balancer health monitor from GET to CONNECT
  cloudscale_ch.cloud.load_balancer_health_monitor:
    uuid: '{{ load_balancer_health_monitor_http.uuid }}'
    delay_s: 2
    timeout_s: 1
    up_threshold: 2
    down_threshold: 3
    type: 'http'
    http:
      expected_codes:
        - 200
        - 202
      method: 'CONNECT'
      url_path: '/'
      version: '1.1'
      host: 'host1'
    tags:
      project: ansible-test
      stage: production
      sla: 24-7
    api_token: xxxxxx
  register: load_balancer_health_monitor
'''

RETURN = '''
href:
  description: API URL to get details about this load balancer health monitor
  returned: success when not state == absent
  type: str
  sample: https://api.cloudscale.ch/v1/load-balancers/health-monitors/ee4952d4-2eba-4dec-8957-7911b3ce245b
uuid:
  description: The unique identifier for this load balancer health monitor
  returned: success
  type: str
  sample: ee4952d4-2eba-4dec-8957-7911b3ce245b
created_at:
  description: The creation date and time of the load balancer health monitor
  returned: success when not state == absent
  type: str
  sample: "2023-02-22T09:55:38.285018Z"
pool:
  description: The pool of the health monitor
  returned: success when not state == absent
  type: dict
  sample: [
            "href": "https://api.cloudscale.ch/v1/load-balancers/pools/618a6cc8-d757-4fab-aa10-d49dc47e667b",
            "uuid": "618a6cc8-d757-4fab-aa10-d49dc47e667b",
            "name": "swimming pool"
          ]
delay_s:
  description: The delay between two successive checks in seconds
  returned: success when not state == absent
  type: int
  sample: 2
timeout_s:
  description: The maximum time allowed for an individual check in seconds
  returned: success when not state == absent
  type: int
  sample: 1
up_threshold:
  description: The number of checks that need to be successful before the monitor_status of a pool member changes to "up"
  returned: success when not state == absent
  type: int
  sample: 2
down_threshold:
  description: The number of checks that need to fail before the monitor_status of a pool member changes to "down"
  returned: success when not state == absent
  type: int
  sample: 3
type:
  description: The type of the health monitor
  returned: success when not state == absent
  type: str
http:
  description: Advanced options for health monitors with type "http" or "https"
  returned: success when not state == absent
  type: dict
  sample: [ {
                "expected_codes": [
                    "200"
                ],
                "method": "GET",
                "url_path": "/",
                "version": "1.0",
                "host": null
            } ]
tags:
  description: Tags assosiated with the load balancer
  returned: success
  type: dict
  sample: { 'project': 'my project' }
'''

from ansible.module_utils.basic import AnsibleModule
from ..module_utils.api import (
    AnsibleCloudscaleBase,
    cloudscale_argument_spec,
)

ALLOWED_STATES = ('present',
                  'absent',
                  )
ALLOWED_HTTP_POST_PARAMS = ('expected_codes',
                            'host',
                            'method',
                            'url_path')


class AnsibleCloudscaleLoadBalancerHealthMonitor(AnsibleCloudscaleBase):

    def __init__(self, module):
        super(AnsibleCloudscaleLoadBalancerHealthMonitor, self).__init__(
            module,
            resource_name='load-balancers/health-monitors',
            resource_key_name='pool',
            resource_create_param_keys=[
                'pool',
                'delay_s',
                'timeout_s',
                'up_threshold',
                'down_threshold',
                'type',
                'http',
                'tags',
            ],
            resource_update_param_keys=[
                'delay_s',
                'timeout_s',
                'up_threshold',
                'down_threshold',
                'expected_codes',
                'http',
                'tags',
            ],
        )

    def query(self):
        # Initialize
        self._resource_data = self.init_resource()

        resource_key_pool = 'pool'
        uuid = self._module.params[self.resource_key_uuid]
        pool = self._module.params[resource_key_pool]
        matching = []

        # Either search by given health monitor's UUID or
        # search the health monitor by its acossiated pool UUID (1:1)
        if uuid is not None:
            super().query()
        else:
            pool = self._module.params[resource_key_pool]
            if pool is not None:

                resources = self._get('%s' % (self.resource_name))

                if resources:
                    for health_monitor in resources:
                        if health_monitor[resource_key_pool]['uuid'] == pool:
                            matching.append(health_monitor)

            # Fail on more than one resource with identical name
            if len(matching) > 1:
                self._module.fail_json(
                    msg="More than one %s resource for pool '%s' exists." % (
                        self.resource_name,
                        resource_key_pool
                    )
                )
            elif len(matching) == 1:
                self._resource_data = matching[0]
                self._resource_data['state'] = "present"

        return self.pre_transform(self._resource_data)

    def update(self, resource):
        updated = False
        for param in self.resource_update_param_keys:
            if param == 'http' and self._module.params.get('http') is not None:
                for subparam in ALLOWED_HTTP_POST_PARAMS:
                    updated = self._http_param_updated(subparam, resource) or updated
            else:
                updated = self._param_updated(param, resource) or updated

        # Refresh if resource was updated in live mode
        if updated and not self._module.check_mode:
            resource = self.query()
        return resource

    def _http_param_updated(self, key, resource):
        param_http = self._module.params.get('http')
        param = param_http[key]

        if param is None:
            return False

        if not resource or key not in resource['http']:
            return False

        is_different = self.find_http_difference(key, resource, param)

        if is_different:
            self._result['changed'] = True

            patch_data = {
                'http': {
                    key: param
                }
            }

            before_data = {
                'http': {
                    key: resource['http'][key]
                }
            }

            self._result['diff']['before'].update(before_data)
            self._result['diff']['after'].update(patch_data)

            if not self._module.check_mode:
                href = resource.get('href')
                if not href:
                    self._module.fail_json(msg='Unable to update %s, no href found.' % key)

                self._patch(href, patch_data)
                return True
        return False

    def find_http_difference(self, key, resource, param):
        is_different = False

        if param != resource['http'][key]:
            is_different = True

        return is_different


def main():
    argument_spec = cloudscale_argument_spec()
    argument_spec.update(dict(
        uuid=dict(type='str'),
        pool=dict(type='str'),
        delay_s=dict(type='int'),
        timeout_s=dict(type='int'),
        up_threshold=dict(type='int'),
        down_threshold=dict(type='int'),
        type=dict(type='str'),
        http=dict(
            type='dict',
            options=dict(
                expected_codes=dict(type='list', elements='str'),
                method=dict(type='str'),
                url_path=dict(type='str'),
                version=dict(type='str'),
                host=dict(type='str'),
            )
        ),
        tags=dict(type='dict'),
        state=dict(default='present', choices=ALLOWED_STATES),
    ))

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=(),
        required_one_of=(),
        required_if=(),
        supports_check_mode=True,
    )

    cloudscale_load_balancer_health_monitor = AnsibleCloudscaleLoadBalancerHealthMonitor(module)
    cloudscale_load_balancer_health_monitor.query_constraint_keys = []

    if module.params['state'] == "absent":
        result = cloudscale_load_balancer_health_monitor.absent()
    else:
        result = cloudscale_load_balancer_health_monitor.present()
    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2021 Praveen Ghuge (@praveenghuge), Karl Dasan (@ikarldasan)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_ddosprotectionplan_info
version_added: "1.7.0"
short_description: Get Azure DDoS protection plan
description:
    - Get facts of Azure DDoS protection plan.
options:
    resource_group:
        description:
            - The name of the resource group.
        type: str
    name:
        description:
            - The name of the DDoS protection plan.
        type: str
extends_documentation_fragment:
- azure.azcollection.azure
author:
    - Praveen Ghuge (@praveenghuge)
    - Karl Dasan (@ikarldasan)
'''


EXAMPLES = '''
- name: Get facts of specific DDoS protection plan
  azure_rm_ddosprotectionplan_info:
    resource_group: myResourceGroup
    name: myDDoSProtectionPlan
'''

RETURN = '''
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureDDoSProtectionPlanInfo(AzureRMModuleBase):
    def __init__(self):
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str'
            ),
            name=dict(
                type='str'
            )
        )
        # store the results of the module operation
        self.results = dict(
            changed=False)
        self.resource_group = None
        self.name = None
        self.tags = None

        super(AzureDDoSProtectionPlanInfo, self).__init__(
            self.module_arg_spec, supports_check_mode=True, supports_tags=False)

    def exec_module(self, **kwargs):

        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        if self.name is not None:
            results = self.get()
        elif self.resource_group:
            # all the DDoS protection plan listed in that specific resource group
            results = self.list_resource_group()
        else:
            # all the DDoS protection plan listed in the subscription
            results = self.list_subscription()

        self.results['ddosprotectionplan'] = [
            self.ddos_protection_plan_to_dict(x) for x in results]
        return self.results

    def get(self):
        response = None
        results = []
        try:
            response = self.network_client.ddos_protection_plans.get(
                self.resource_group, self.name)
            self.log("Response : {0}".format(response))
        except ResourceNotFoundError as e:
            self.fail('Could not get info for DDoS protection plan. {0}'.format(str(e)))

        if response and self.has_tags(response.tags, self.tags):
            results = [response]
        return results

    def list_resource_group(self):
        self.log('List items for resource group')
        try:
            response = self.network_client.ddos_protection_plans.list_by_resource_group(
                self.resource_group)

        except ResourceNotFoundError as exc:
            self.fail(
                "Failed to list for resource group {0} - {1}".format(self.resource_group, str(exc)))

        results = []
        for item in response:
            if self.has_tags(item.tags, self.tags):
                results.append(item)
        return results

    def list_subscription(self):
        self.log('List items for subscription')
        try:
            response = self.network_client.ddos_protection_plans.list()

        except ResourceNotFoundError as exc:
            self.fail(
                "Failed to list DDoS protection plan in the subscription - {0}".format(str(exc)))

        results = []
        for item in response:
            if self.has_tags(item.tags, self.tags):
                results.append(item)
        return results

    def ddos_protection_plan_to_dict(self, item):
        # turn DDoS protection plan object into a dictionary (serialization)
        ddos_protection_plan = item.as_dict()

        result = dict(
            additional_properties=ddos_protection_plan.get('additional_properties', None),
            id=ddos_protection_plan.get('id', None),
            name=ddos_protection_plan.get('name', None),
            type=ddos_protection_plan.get('type', None),
            location=ddos_protection_plan.get('location', None),
            tags=ddos_protection_plan.get('tags', None),
            etag=ddos_protection_plan.get('etag', None),
            resource_guid=ddos_protection_plan.get('resource_guid', None),
            provisioning_state=ddos_protection_plan.get('provisioning_state', None),
            virtual_networks=ddos_protection_plan.get('virtual_networks', None)
        )
        return result


def main():
    AzureDDoSProtectionPlanInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2025, mariusbertram <marius@brtrm.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_user_execute_actions_email

short_description: Send a Keycloak execute-actions email to a user

version_added: 12.0.0

description:
  - Triggers the Keycloak endpoint C(execute-actions-email) for a user.
    This sends an email with one or more required actions the user must complete (for example resetting the password).
  - If no O(actions) list is provided, the default action C(UPDATE_PASSWORD) is used.
  - You must supply either the user's O(id) or O(username). Supplying only C(username) causes an extra lookup call.
  - This module always reports RV(ignore:changed=true) because sending an email is a side effect and cannot be made idempotent.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  auth_username:
    aliases: []
  realm:
    description:
      - The Keycloak realm where the user resides.
    type: str
    default: master
  id:
    description:
      - The unique ID (UUID) of the user.
      - Mutually exclusive with O(username).
    type: str
  username:
    description:
      - Username of the user.
      - Mutually exclusive with O(id).
    type: str
  actions:
    description:
      - List of required actions to include in the email.
    type: list
    elements: str
    default:
      - UPDATE_PASSWORD
  client_id:
    description:
      - Optional client ID used for the redirect link.
    aliases: [clientId]
    type: str
  redirect_uri:
    description:
      - Optional redirect URI. Must be valid for the given client if O(client_id) is set.
    aliases: [redirectUri]
    type: str
  lifespan:
    description:
      - Optional lifespan (in seconds) for the action token (supported on newer Keycloak versions). Forwarded as query parameter if provided.
    type: int
extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes
author:
  - Marius Bertram (@mariusbertram)
"""

EXAMPLES = r"""
- name: Password reset email (default action) with 1h lifespan
  community.general.keycloak_user_execute_actions_email:
    username: johndoe
    realm: MyRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: ADMIN
    auth_password: SECRET
    lifespan: 3600
  delegate_to: localhost

- name: Multiple required actions using token auth
  community.general.keycloak_user_execute_actions_email:
    username: johndoe
    actions:
      - UPDATE_PASSWORD
      - VERIFY_EMAIL
    realm: MyRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    token: TOKEN
  delegate_to: localhost

- name: Email by user id with redirect
  community.general.keycloak_user_execute_actions_email:
    id: 9d59aa76-2755-48c6-b1af-beb70a82c3cd
    client_id: my-frontend
    redirect_uri: https://app.example.com/post-actions
    actions:
      - UPDATE_PASSWORD
    realm: MyRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: ADMIN
    auth_password: SECRET
  delegate_to: localhost
"""

RETURN = r"""
user_id:
  description: The user ID the email was (or would be, in check mode) sent to.
  returned: success
  type: str
actions:
  description: List of actions included in the email.
  returned: success
  type: list
  elements: str
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    get_token,
    keycloak_argument_spec,
)


def main():
    argument_spec = keycloak_argument_spec()
    # Avoid alias collision as in keycloak_user: clear auth_username aliases locally
    argument_spec["auth_username"]["aliases"] = []

    meta_args = dict(
        realm=dict(type="str", default="master"),
        id=dict(type="str"),
        username=dict(type="str"),
        actions=dict(type="list", elements="str", default=["UPDATE_PASSWORD"]),
        client_id=dict(type="str", aliases=["clientId"]),
        redirect_uri=dict(type="str", aliases=["redirectUri"]),
        lifespan=dict(type="int"),
    )
    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=[["id", "username"]],
        mutually_exclusive=[["id", "username"]],
    )

    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params.get("realm")
    user_id = module.params.get("id")
    username = module.params.get("username")
    actions = module.params.get("actions")
    client_id = module.params.get("client_id")
    redirect_uri = module.params.get("redirect_uri")
    lifespan = module.params.get("lifespan")

    # Resolve user ID if only username is provided
    if user_id is None:
        user_obj = kc.get_user_by_username(username=username, realm=realm)
        if user_obj is None:
            module.fail_json(msg=f"User '{username}' not found in realm {realm}")
        user_id = user_obj["id"]

    if module.check_mode:
        module.exit_json(
            changed=True, msg=f"Would send execute-actions email to user {user_id}", user_id=user_id, actions=actions
        )

    try:
        kc.send_execute_actions_email(
            user_id=user_id,
            realm=realm,
            client_id=client_id,
            data=actions,
            redirect_uri=redirect_uri,
            lifespan=lifespan,
        )
    except Exception as e:
        module.fail_json(msg=str(e))

    module.exit_json(
        changed=True, msg=f"Execute-actions email sent to user {user_id}", user_id=user_id, actions=actions
    )


if __name__ == "__main__":
    main()



# Copyright (c) 2023, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this plugin util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t
from subprocess import PIPE, Popen

from ansible.module_utils.common.process import get_bin_path
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.crypto.plugins.module_utils._gnupg.cli import (
    GPGError,
    GPGRunner,
)


class PluginGPGRunner(GPGRunner):
    def __init__(
        self, *, executable: str | None = None, cwd: str | None = None
    ) -> None:
        if executable is None:
            try:
                executable = get_bin_path("gpg")
            except ValueError as exc:
                raise GPGError(
                    "Cannot find the `gpg` executable on the controller"
                ) from exc
        self.executable = executable
        self.cwd = cwd

    def run_command(
        self, command: list[str], *, check_rc: bool = True, data: bytes | None = None
    ) -> tuple[int, str, str]:
        """
        Run ``[gpg] + command`` and return ``(rc, stdout, stderr)``.

        If ``data`` is not ``None``, it will be provided as stdin.
        The code assumes it is a bytes string.

        Returned stdout and stderr are native Python strings.
        Pass ``check_rc=False`` to allow return codes != 0.

        Raises a ``GPGError`` in case of errors.
        """
        command = [self.executable] + command
        with Popen(
            command, shell=False, cwd=self.cwd, stdin=PIPE, stdout=PIPE, stderr=PIPE
        ) as p:
            stdout, stderr = p.communicate(input=data)
            stdout_n = to_text(stdout, errors="surrogate_or_replace")
            stderr_n = to_text(stderr, errors="surrogate_or_replace")
            if check_rc and p.returncode != 0:
                raise GPGError(
                    f'Running {" ".join(command)} yielded return code {p.returncode} with stdout: "{stdout_n}" and stderr: "{stderr_n}")'
                )
            return t.cast(int, p.returncode), stdout_n, stderr_n


__all__ = ("PluginGPGRunner",)



# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import copy
from typing import Optional

try:
    import botocore.exceptions as exception_waiter
    import botocore.waiter as core_waiter
except ImportError:
    pass  # caught by HAS_BOTO3

from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import RetryingBotoClientWrapper

ec2_data = {
    "version": 2,
    "waiters": {
        "ImageAvailable": {
            "operation": "DescribeImages",
            "maxAttempts": 80,
            "delay": 15,
            "acceptors": [
                {"state": "success", "matcher": "pathAll", "argument": "Images[].State", "expected": "available"},
                {"matcher": "error", "expected": "InvalidAMIID.NotFound", "state": "retry"},
                {"state": "failure", "matcher": "pathAny", "argument": "Images[].State", "expected": "failed"},
            ],
        },
        "InternetGatewayExists": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeInternetGateways",
            "acceptors": [
                {"matcher": "path", "expected": True, "argument": "length(InternetGateways) > `0`", "state": "success"},
                {"matcher": "error", "expected": "InvalidInternetGatewayID.NotFound", "state": "retry"},
            ],
        },
        "InternetGatewayAttached": {
            "operation": "DescribeInternetGateways",
            "delay": 5,
            "maxAttempts": 40,
            "acceptors": [
                {
                    "expected": "available",
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "InternetGateways[].Attachments[].State",
                },
                {"matcher": "error", "expected": "InvalidInternetGatewayID.NotFound", "state": "retry"},
            ],
        },
        "TransitGatewayAvailable": {
            "operation": "DescribeTransitGateways",
            "delay": 5,
            "maxAttempts": 120,
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "TransitGateways[].State",
                    "expected": "available",
                },
                {
                    "state": "retry",
                    "matcher": "pathAll",
                    "argument": "TransitGateways[].State",
                    "expected": "pending",
                },
                {
                    "state": "failure",
                    "matcher": "pathAny",
                    "argument": "TransitGateways[].State",
                    "expected": "failed",
                },
            ],
        },
        "TransitGatewayDeleted": {
            "operation": "DescribeTransitGateways",
            "delay": 5,
            "maxAttempts": 120,
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "TransitGateways[].State",
                    "expected": "deleted",
                },
                {
                    "state": "retry",
                    "matcher": "pathAll",
                    "argument": "TransitGateways[].State",
                    "expected": "deleting",
                },
                {
                    "state": "failure",
                    "matcher": "pathAny",
                    "argument": "TransitGateways[].State",
                    "expected": "failed",
                },
            ],
        },
        "TGWVpcAttachmentAvailable": {
            "operation": "DescribeTransitGatewayVpcAttachments",
            "delay": 5,
            "maxAttempts": 120,
            "acceptors": [
                {
                    "expected": "available",
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "TransitGatewayVpcAttachments[].State",
                },
                {
                    "state": "retry",
                    "matcher": "pathAny",
                    "argument": "TransitGatewayVpcAttachments[].State",
                    "expected": "pending",
                },
                {"matcher": "error", "expected": "InvalidRouteTableID.NotFound", "state": "retry"},
            ],
        },
        "TGWVpcAttachmentDeleted": {
            "operation": "DescribeTransitGatewayVpcAttachments",
            "delay": 5,
            "maxAttempts": 120,
            "acceptors": [
                {
                    "state": "retry",
                    "matcher": "pathAll",
                    "argument": "TransitGatewayVpcAttachments[].State",
                    "expected": "deleting",
                },
                {
                    "state": "success",
                    "expected": "deleted",
                    "matcher": "pathAll",
                    "argument": "TransitGatewayVpcAttachments[].State",
                },
                {
                    "expected": True,
                    "matcher": "path",
                    "state": "success",
                    "argument": "length(TransitGatewayVpcAttachments[]) == `0`",
                },
                {"matcher": "error", "expected": "InvalidRouteTableID.NotFound", "state": "retry"},
            ],
        },
        "NetworkInterfaceAttached": {
            "operation": "DescribeNetworkInterfaces",
            "delay": 5,
            "maxAttempts": 40,
            "acceptors": [
                {
                    "expected": "attached",
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "NetworkInterfaces[].Attachment.Status",
                },
                {"expected": "InvalidNetworkInterfaceID.NotFound", "matcher": "error", "state": "failure"},
            ],
        },
        "NetworkInterfaceAvailable": {
            "operation": "DescribeNetworkInterfaces",
            "delay": 5,
            "maxAttempts": 40,
            "acceptors": [
                {
                    "expected": "available",
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "NetworkInterfaces[].Status",
                },
                {"expected": "InvalidNetworkInterfaceID.NotFound", "matcher": "error", "state": "retry"},
            ],
        },
        "NetworkInterfaceDeleted": {
            "operation": "DescribeNetworkInterfaces",
            "delay": 5,
            "maxAttempts": 40,
            "acceptors": [
                {
                    "matcher": "path",
                    "expected": True,
                    "argument": "length(NetworkInterfaces[]) > `0`",
                    "state": "retry",
                },
                {
                    "matcher": "path",
                    "expected": True,
                    "argument": "length(NetworkInterfaces[]) == `0`",
                    "state": "success",
                },
                {"expected": "InvalidNetworkInterfaceID.NotFound", "matcher": "error", "state": "success"},
            ],
        },
        "NetworkInterfaceDeleteOnTerminate": {
            "operation": "DescribeNetworkInterfaces",
            "delay": 5,
            "maxAttempts": 10,
            "acceptors": [
                {
                    "expected": True,
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "NetworkInterfaces[].Attachment.DeleteOnTermination",
                },
                {"expected": "InvalidNetworkInterfaceID.NotFound", "matcher": "error", "state": "failure"},
            ],
        },
        "NetworkInterfaceNoDeleteOnTerminate": {
            "operation": "DescribeNetworkInterfaces",
            "delay": 5,
            "maxAttempts": 10,
            "acceptors": [
                {
                    "expected": False,
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "NetworkInterfaces[].Attachment.DeleteOnTermination",
                },
                {"expected": "InvalidNetworkInterfaceID.NotFound", "matcher": "error", "state": "failure"},
            ],
        },
        "RouteTableExists": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeRouteTables",
            "acceptors": [
                {"matcher": "path", "expected": True, "argument": "length(RouteTables[]) > `0`", "state": "success"},
                {"matcher": "error", "expected": "InvalidRouteTableID.NotFound", "state": "retry"},
            ],
        },
        "SecurityGroupExists": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeSecurityGroups",
            "acceptors": [
                {"matcher": "path", "expected": True, "argument": "length(SecurityGroups[]) > `0`", "state": "success"},
                {"matcher": "error", "expected": "InvalidGroup.NotFound", "state": "retry"},
            ],
        },
        "SnapshotCompleted": {
            "delay": 15,
            "operation": "DescribeSnapshots",
            "maxAttempts": 40,
            "acceptors": [
                {"expected": "completed", "matcher": "pathAll", "state": "success", "argument": "Snapshots[].State"}
            ],
        },
        "SubnetAvailable": {
            "delay": 15,
            "operation": "DescribeSubnets",
            "maxAttempts": 40,
            "acceptors": [
                {"expected": "available", "matcher": "pathAll", "state": "success", "argument": "Subnets[].State"}
            ],
        },
        "SubnetExists": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeSubnets",
            "acceptors": [
                {"matcher": "path", "expected": True, "argument": "length(Subnets[]) > `0`", "state": "success"},
                {"matcher": "error", "expected": "InvalidSubnetID.NotFound", "state": "retry"},
            ],
        },
        "SubnetHasMapPublic": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeSubnets",
            "acceptors": [
                {
                    "matcher": "pathAll",
                    "expected": True,
                    "argument": "Subnets[].MapPublicIpOnLaunch",
                    "state": "success",
                },
            ],
        },
        "SubnetNoMapPublic": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeSubnets",
            "acceptors": [
                {
                    "matcher": "pathAll",
                    "expected": False,
                    "argument": "Subnets[].MapPublicIpOnLaunch",
                    "state": "success",
                },
            ],
        },
        "SubnetHasAssignIpv6": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeSubnets",
            "acceptors": [
                {
                    "matcher": "pathAll",
                    "expected": True,
                    "argument": "Subnets[].AssignIpv6AddressOnCreation",
                    "state": "success",
                },
            ],
        },
        "SubnetNoAssignIpv6": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeSubnets",
            "acceptors": [
                {
                    "matcher": "pathAll",
                    "expected": False,
                    "argument": "Subnets[].AssignIpv6AddressOnCreation",
                    "state": "success",
                },
            ],
        },
        "SubnetDeleted": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeSubnets",
            "acceptors": [
                {"matcher": "path", "expected": True, "argument": "length(Subnets[]) > `0`", "state": "retry"},
                {"matcher": "error", "expected": "InvalidSubnetID.NotFound", "state": "success"},
            ],
        },
        "VpcAvailable": {
            "delay": 15,
            "operation": "DescribeVpcs",
            "maxAttempts": 40,
            "acceptors": [
                {"expected": "available", "matcher": "pathAll", "state": "success", "argument": "Vpcs[].State"}
            ],
        },
        "VpcExists": {
            "operation": "DescribeVpcs",
            "delay": 1,
            "maxAttempts": 5,
            "acceptors": [
                {"matcher": "status", "expected": 200, "state": "success"},
                {"matcher": "error", "expected": "InvalidVpcID.NotFound", "state": "retry"},
            ],
        },
        "VpcEndpointExists": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeVpcEndpoints",
            "acceptors": [
                {"matcher": "path", "expected": True, "argument": "length(VpcEndpoints[]) > `0`", "state": "success"},
                {"matcher": "error", "expected": "InvalidVpcEndpointId.NotFound", "state": "retry"},
            ],
        },
        "VpnGatewayExists": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeVpnGateways",
            "acceptors": [
                {"matcher": "path", "expected": True, "argument": "length(VpnGateways[]) > `0`", "state": "success"},
                {"matcher": "error", "expected": "InvalidVpnGatewayID.NotFound", "state": "retry"},
            ],
        },
        "VpnGatewayDetached": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeVpnGateways",
            "acceptors": [
                {
                    "expected": "detached",
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "VpnGateways[].VpcAttachments[].State",
                },
            ],
        },
        "VpnGatewayAttached": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeVpnGateways",
            "acceptors": [
                {
                    "expected": "attached",
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "VpnGateways[].VpcAttachments[].State",
                },
            ],
        },
        "NatGatewayDeleted": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeNatGateways",
            "acceptors": [
                {"state": "success", "matcher": "pathAll", "expected": "deleted", "argument": "NatGateways[].State"},
                {"state": "success", "matcher": "error", "expected": "NatGatewayNotFound"},
            ],
        },
        "NatGatewayAvailable": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeNatGateways",
            "acceptors": [
                {"state": "success", "matcher": "pathAll", "expected": "available", "argument": "NatGateways[].State"},
                {"state": "retry", "matcher": "error", "expected": "NatGatewayNotFound"},
            ],
        },
        "VolumeAvailable": {
            "delay": 15,
            "maxAttempts": 40,
            "operation": "DescribeVolumes",
            "acceptors": [
                {"state": "success", "matcher": "pathAll", "expected": "available", "argument": "Volumes[].State"},
                {"state": "failure", "matcher": "pathAny", "argument": "Volumes[].State", "expected": "deleted"},
            ],
        },
        "VolumeDeleted": {
            "delay": 15,
            "maxAttempts": 40,
            "operation": "DescribeVolumes",
            "acceptors": [
                {"state": "success", "matcher": "pathAll", "expected": "deleted", "argument": "Volumes[].State"},
                {"state": "success", "matcher": "error", "expected": "InvalidVolume.NotFound"},
            ],
        },
        "VolumeInUse": {
            "delay": 15,
            "maxAttempts": 40,
            "operation": "DescribeVolumes",
            "acceptors": [
                {"state": "success", "matcher": "pathAll", "expected": "in-use", "argument": "Volumes[].State"},
                {"state": "failure", "matcher": "pathAny", "argument": "Volumes[].State", "expected": "deleted"},
            ],
        },
    },
}


waf_data = {
    "version": 2,
    "waiters": {
        "ChangeTokenInSync": {
            "delay": 20,
            "maxAttempts": 60,
            "operation": "GetChangeTokenStatus",
            "acceptors": [
                {"matcher": "path", "expected": True, "argument": "ChangeTokenStatus == 'INSYNC'", "state": "success"},
                {"matcher": "error", "expected": "WAFInternalErrorException", "state": "retry"},
            ],
        }
    },
}

eks_data = {
    "version": 2,
    "waiters": {
        "ClusterActive": {
            "delay": 20,
            "maxAttempts": 60,
            "operation": "DescribeCluster",
            "acceptors": [
                {"state": "success", "matcher": "path", "argument": "cluster.status", "expected": "ACTIVE"},
                {"state": "retry", "matcher": "error", "expected": "ResourceNotFoundException"},
            ],
        },
        "ClusterDeleted": {
            "delay": 20,
            "maxAttempts": 60,
            "operation": "DescribeCluster",
            "acceptors": [
                {"state": "retry", "matcher": "path", "argument": "cluster.status != 'DELETED'", "expected": True},
                {"state": "success", "matcher": "error", "expected": "ResourceNotFoundException"},
            ],
        },
        "FargateProfileActive": {
            "delay": 20,
            "maxAttempts": 30,
            "operation": "DescribeFargateProfile",
            "acceptors": [
                {"state": "success", "matcher": "path", "argument": "fargateProfile.status", "expected": "ACTIVE"},
                {"state": "retry", "matcher": "error", "expected": "ResourceNotFoundException"},
            ],
        },
        "FargateProfileDeleted": {
            "delay": 20,
            "maxAttempts": 30,
            "operation": "DescribeFargateProfile",
            "acceptors": [
                {
                    "state": "retry",
                    "matcher": "path",
                    "argument": "fargateProfile.status == 'DELETING'",
                    "expected": True,
                },
                {"state": "success", "matcher": "error", "expected": "ResourceNotFoundException"},
            ],
        },
        "NodegroupActive": {
            "delay": 20,
            "maxAttempts": 60,
            "operation": "DescribeNodegroup",
            "acceptors": [
                {"state": "success", "matcher": "path", "argument": "nodegroup.status", "expected": "ACTIVE"},
                {"state": "retry", "matcher": "error", "expected": "ResourceNotFoundException"},
            ],
        },
        "NodegroupDeleted": {
            "delay": 20,
            "maxAttempts": 60,
            "operation": "DescribeNodegroup",
            "acceptors": [
                {"state": "retry", "matcher": "path", "argument": "nodegroup.status == 'DELETING'", "expected": True},
                {"state": "success", "matcher": "error", "expected": "ResourceNotFoundException"},
            ],
        },
    },
}


elb_data = {
    "version": 2,
    "waiters": {
        "AnyInstanceInService": {
            "acceptors": [
                {
                    "argument": "InstanceStates[].State",
                    "expected": "InService",
                    "matcher": "pathAny",
                    "state": "success",
                }
            ],
            "delay": 15,
            "maxAttempts": 40,
            "operation": "DescribeInstanceHealth",
        },
        "InstanceDeregistered": {
            "delay": 15,
            "operation": "DescribeInstanceHealth",
            "maxAttempts": 40,
            "acceptors": [
                {
                    "expected": "OutOfService",
                    "matcher": "pathAll",
                    "state": "success",
                    "argument": "InstanceStates[].State",
                },
                {"matcher": "error", "expected": "InvalidInstance", "state": "success"},
            ],
        },
        "InstanceInService": {
            "acceptors": [
                {
                    "argument": "InstanceStates[].State",
                    "expected": "InService",
                    "matcher": "pathAll",
                    "state": "success",
                },
                {"matcher": "error", "expected": "InvalidInstance", "state": "retry"},
            ],
            "delay": 15,
            "maxAttempts": 40,
            "operation": "DescribeInstanceHealth",
        },
        "LoadBalancerCreated": {
            "delay": 10,
            "maxAttempts": 60,
            "operation": "DescribeLoadBalancers",
            "acceptors": [
                {
                    "matcher": "path",
                    "expected": True,
                    "argument": "length(LoadBalancerDescriptions[]) > `0`",
                    "state": "success",
                },
                {
                    "matcher": "error",
                    "expected": "LoadBalancerNotFound",
                    "state": "retry",
                },
            ],
        },
        "LoadBalancerDeleted": {
            "delay": 10,
            "maxAttempts": 60,
            "operation": "DescribeLoadBalancers",
            "acceptors": [
                {
                    "matcher": "path",
                    "expected": True,
                    "argument": "length(LoadBalancerDescriptions[]) > `0`",
                    "state": "retry",
                },
                {
                    "matcher": "error",
                    "expected": "LoadBalancerNotFound",
                    "state": "success",
                },
            ],
        },
    },
}

elbv2_data = {
    "version": 2,
    "waiters": {
        "LoadBalancerAvailable": {
            "delay": 15,
            "operation": "DescribeLoadBalancers",
            "maxAttempts": 40,
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "LoadBalancers[].State.Code",
                    "expected": "active",
                },
                {
                    "state": "retry",
                    "matcher": "pathAny",
                    "argument": "LoadBalancers[].State.Code",
                    "expected": "provisioning",
                },
                {"state": "retry", "matcher": "error", "expected": "LoadBalancerNotFound"},
            ],
        },
        "LoadBalancerIpAddressTypeIpv4": {
            "delay": 15,
            "operation": "DescribeLoadBalancers",
            "maxAttempts": 40,
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "LoadBalancers[].IpAddressType",
                    "expected": "ipv4",
                },
                {
                    "state": "retry",
                    "matcher": "pathAny",
                    "argument": "LoadBalancers[].IpAddressType",
                    "expected": "dualstack",
                },
                {"state": "failure", "matcher": "error", "expected": "LoadBalancerNotFound"},
            ],
        },
        "LoadBalancerIpAddressTypeDualStack": {
            "delay": 15,
            "operation": "DescribeLoadBalancers",
            "maxAttempts": 40,
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "LoadBalancers[].IpAddressType",
                    "expected": "dualstack",
                },
                {
                    "state": "retry",
                    "matcher": "pathAny",
                    "argument": "LoadBalancers[].IpAddressType",
                    "expected": "ipv4",
                },
                {"state": "failure", "matcher": "error", "expected": "LoadBalancerNotFound"},
            ],
        },
        "LoadBalancersDeleted": {
            "delay": 15,
            "operation": "DescribeLoadBalancers",
            "maxAttempts": 40,
            "acceptors": [
                {
                    "state": "retry",
                    "matcher": "pathAll",
                    "argument": "LoadBalancers[].State.Code",
                    "expected": "active",
                },
                {"matcher": "error", "expected": "LoadBalancerNotFound", "state": "success"},
            ],
        },
    },
}


rds_data = {
    "version": 2,
    "waiters": {
        "DBClusterPromoting": {
            "delay": 5,
            "maxAttempts": 60,
            "operation": "DescribeDBClusters",
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "DBClusters[].Status",
                    "expected": "promoting",
                },
            ],
        },
        "DBInstanceStopped": {
            "delay": 20,
            "maxAttempts": 60,
            "operation": "DescribeDBInstances",
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "DBInstances[].DBInstanceStatus",
                    "expected": "stopped",
                },
            ],
        },
        "DBClusterAvailable": {
            "delay": 20,
            "maxAttempts": 60,
            "operation": "DescribeDBClusters",
            "acceptors": [
                {"state": "success", "matcher": "pathAll", "argument": "DBClusters[].Status", "expected": "available"},
                {"state": "retry", "matcher": "error", "expected": "DBClusterNotFoundFault"},
            ],
        },
        "DBClusterDeleted": {
            "delay": 20,
            "maxAttempts": 60,
            "operation": "DescribeDBClusters",
            "acceptors": [
                {"state": "success", "matcher": "pathAll", "argument": "DBClusters[].Status", "expected": "stopped"},
                {"state": "success", "matcher": "error", "expected": "DBClusterNotFoundFault"},
            ],
        },
        "ReadReplicaPromoted": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeDBInstances",
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "path",
                    "argument": "length(DBInstances[].StatusInfos) == `0`",
                    "expected": True,
                },
                {
                    "state": "retry",
                    "matcher": "pathAny",
                    "argument": "DBInstances[].StatusInfos[].Status",
                    "expected": "replicating",
                },
            ],
        },
        "RoleAssociated": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeDBInstances",
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "DBInstances[].AssociatedRoles[].Status",
                    "expected": "ACTIVE",
                },
                {
                    "state": "retry",
                    "matcher": "pathAny",
                    "argument": "DBInstances[].AssociatedRoles[].Status",
                    "expected": "PENDING",
                },
            ],
        },
        "RoleDisassociated": {
            "delay": 5,
            "maxAttempts": 40,
            "operation": "DescribeDBInstances",
            "acceptors": [
                {
                    "state": "success",
                    "matcher": "pathAll",
                    "argument": "DBInstances[].AssociatedRoles[].Status",
                    "expected": "ACTIVE",
                },
                {
                    "state": "retry",
                    "matcher": "pathAny",
                    "argument": "DBInstances[].AssociatedRoles[].Status",
                    "expected": "PENDING",
                },
                {
                    "state": "success",
                    "matcher": "path",
                    "argument": "length(DBInstances[].AssociatedRoles[]) == `0`",
                    "expected": True,
                },
            ],
        },
    },
}


route53_data = {
    "version": 2,
    "waiters": {
        "ResourceRecordSetsChanged": {
            "delay": 30,
            "maxAttempts": 60,
            "operation": "GetChange",
            "acceptors": [
                {"matcher": "path", "expected": "INSYNC", "argument": "ChangeInfo.Status", "state": "success"}
            ],
        }
    },
}


def _inject_limit_retries(model):
    extra_retries = [
        "RequestLimitExceeded",
        "Unavailable",
        "ServiceUnavailable",
        "InternalFailure",
        "InternalError",
        "TooManyRequestsException",
        "Throttling",
    ]

    acceptors = []
    for error in extra_retries:
        acceptors.append({"state": "success", "matcher": "error", "expected": error})

    _model = copy.deepcopy(model)

    for waiter in model["waiters"]:
        _model["waiters"][waiter]["acceptors"].extend(acceptors)

    return _model


def ec2_model(name):
    ec2_models = core_waiter.WaiterModel(waiter_config=_inject_limit_retries(ec2_data))
    return ec2_models.get_waiter(name)


def waf_model(name):
    waf_models = core_waiter.WaiterModel(waiter_config=_inject_limit_retries(waf_data))
    return waf_models.get_waiter(name)


def eks_model(name):
    eks_models = core_waiter.WaiterModel(waiter_config=_inject_limit_retries(eks_data))
    return eks_models.get_waiter(name)


def elbv2_model(name):
    elbv2_models = core_waiter.WaiterModel(waiter_config=_inject_limit_retries(elbv2_data))
    return elbv2_models.get_waiter(name)


def elb_model(name):
    elb_models = core_waiter.WaiterModel(waiter_config=_inject_limit_retries(elb_data))
    return elb_models.get_waiter(name)


def rds_model(name):
    rds_models = core_waiter.WaiterModel(waiter_config=_inject_limit_retries(rds_data))
    return rds_models.get_waiter(name)


def route53_model(name):
    route53_models = core_waiter.WaiterModel(waiter_config=_inject_limit_retries(route53_data))
    return route53_models.get_waiter(name)


waiters_by_name = {
    ("EC2", "image_available"): lambda ec2: core_waiter.Waiter(
        "image_available", ec2_model("ImageAvailable"), core_waiter.NormalizedOperationMethod(ec2.describe_images)
    ),
    ("EC2", "internet_gateway_exists"): lambda ec2: core_waiter.Waiter(
        "internet_gateway_exists",
        ec2_model("InternetGatewayExists"),
        core_waiter.NormalizedOperationMethod(ec2.describe_internet_gateways),
    ),
    ("EC2", "internet_gateway_attached"): lambda ec2: core_waiter.Waiter(
        "internet_gateway_attached",
        ec2_model("InternetGatewayAttached"),
        core_waiter.NormalizedOperationMethod(ec2.describe_internet_gateways),
    ),
    ("EC2", "transit_gateway_available"): lambda ec2: core_waiter.Waiter(
        "transit_gateway_available",
        ec2_model("TransitGatewayAvailable"),
        core_waiter.NormalizedOperationMethod(ec2.describe_transit_gateways),
    ),
    ("EC2", "transit_gateway_deleted"): lambda ec2: core_waiter.Waiter(
        "transit_gateway_deleted",
        ec2_model("TransitGatewayDeleted"),
        core_waiter.NormalizedOperationMethod(ec2.describe_transit_gateways),
    ),
    ("EC2", "transit_gateway_vpc_attachment_available"): lambda ec2: core_waiter.Waiter(
        "transit_gateway_vpc_attachment_available",
        ec2_model("TGWVpcAttachmentAvailable"),
        core_waiter.NormalizedOperationMethod(ec2.describe_transit_gateway_vpc_attachments),
    ),
    ("EC2", "transit_gateway_vpc_attachment_deleted"): lambda ec2: core_waiter.Waiter(
        "transit_gateway_vpc_attachment_deleted",
        ec2_model("TGWVpcAttachmentDeleted"),
        core_waiter.NormalizedOperationMethod(ec2.describe_transit_gateway_vpc_attachments),
    ),
    ("EC2", "network_interface_attached"): lambda ec2: core_waiter.Waiter(
        "network_interface_attached",
        ec2_model("NetworkInterfaceAttached"),
        core_waiter.NormalizedOperationMethod(ec2.describe_network_interfaces),
    ),
    ("EC2", "network_interface_deleted"): lambda ec2: core_waiter.Waiter(
        "network_interface_deleted",
        ec2_model("NetworkInterfaceDeleted"),
        core_waiter.NormalizedOperationMethod(ec2.describe_network_interfaces),
    ),
    ("EC2", "network_interface_available"): lambda ec2: core_waiter.Waiter(
        "network_interface_available",
        ec2_model("NetworkInterfaceAvailable"),
        core_waiter.NormalizedOperationMethod(ec2.describe_network_interfaces),
    ),
    ("EC2", "network_interface_delete_on_terminate"): lambda ec2: core_waiter.Waiter(
        "network_interface_delete_on_terminate",
        ec2_model("NetworkInterfaceDeleteOnTerminate"),
        core_waiter.NormalizedOperationMethod(ec2.describe_network_interfaces),
    ),
    ("EC2", "network_interface_no_delete_on_terminate"): lambda ec2: core_waiter.Waiter(
        "network_interface_no_delete_on_terminate",
        ec2_model("NetworkInterfaceNoDeleteOnTerminate"),
        core_waiter.NormalizedOperationMethod(ec2.describe_network_interfaces),
    ),
    ("EC2", "route_table_exists"): lambda ec2: core_waiter.Waiter(
        "route_table_exists",
        ec2_model("RouteTableExists"),
        core_waiter.NormalizedOperationMethod(ec2.describe_route_tables),
    ),
    ("EC2", "security_group_exists"): lambda ec2: core_waiter.Waiter(
        "security_group_exists",
        ec2_model("SecurityGroupExists"),
        core_waiter.NormalizedOperationMethod(ec2.describe_security_groups),
    ),
    ("EC2", "snapshot_completed"): lambda ec2: core_waiter.Waiter(
        "snapshot_completed",
        ec2_model("SnapshotCompleted"),
        core_waiter.NormalizedOperationMethod(ec2.describe_snapshots),
    ),
    ("EC2", "subnet_available"): lambda ec2: core_waiter.Waiter(
        "subnet_available", ec2_model("SubnetAvailable"), core_waiter.NormalizedOperationMethod(ec2.describe_subnets)
    ),
    ("EC2", "subnet_exists"): lambda ec2: core_waiter.Waiter(
        "subnet_exists", ec2_model("SubnetExists"), core_waiter.NormalizedOperationMethod(ec2.describe_subnets)
    ),
    ("EC2", "subnet_has_map_public"): lambda ec2: core_waiter.Waiter(
        "subnet_has_map_public",
        ec2_model("SubnetHasMapPublic"),
        core_waiter.NormalizedOperationMethod(ec2.describe_subnets),
    ),
    ("EC2", "subnet_no_map_public"): lambda ec2: core_waiter.Waiter(
        "subnet_no_map_public",
        ec2_model("SubnetNoMapPublic"),
        core_waiter.NormalizedOperationMethod(ec2.describe_subnets),
    ),
    ("EC2", "subnet_has_assign_ipv6"): lambda ec2: core_waiter.Waiter(
        "subnet_has_assign_ipv6",
        ec2_model("SubnetHasAssignIpv6"),
        core_waiter.NormalizedOperationMethod(ec2.describe_subnets),
    ),
    ("EC2", "subnet_no_assign_ipv6"): lambda ec2: core_waiter.Waiter(
        "subnet_no_assign_ipv6",
        ec2_model("SubnetNoAssignIpv6"),
        core_waiter.NormalizedOperationMethod(ec2.describe_subnets),
    ),
    ("EC2", "subnet_deleted"): lambda ec2: core_waiter.Waiter(
        "subnet_deleted", ec2_model("SubnetDeleted"), core_waiter.NormalizedOperationMethod(ec2.describe_subnets)
    ),
    ("EC2", "vpc_available"): lambda ec2: core_waiter.Waiter(
        "vpc_available", ec2_model("VpcAvailable"), core_waiter.NormalizedOperationMethod(ec2.describe_vpcs)
    ),
    ("EC2", "vpc_exists"): lambda ec2: core_waiter.Waiter(
        "vpc_exists", ec2_model("VpcExists"), core_waiter.NormalizedOperationMethod(ec2.describe_vpcs)
    ),
    ("EC2", "vpc_endpoint_exists"): lambda ec2: core_waiter.Waiter(
        "vpc_endpoint_exists",
        ec2_model("VpcEndpointExists"),
        core_waiter.NormalizedOperationMethod(ec2.describe_vpc_endpoints),
    ),
    ("EC2", "vpn_gateway_exists"): lambda ec2: core_waiter.Waiter(
        "vpn_gateway_exists",
        ec2_model("VpnGatewayExists"),
        core_waiter.NormalizedOperationMethod(ec2.describe_vpn_gateways),
    ),
    ("EC2", "vpn_gateway_detached"): lambda ec2: core_waiter.Waiter(
        "vpn_gateway_detached",
        ec2_model("VpnGatewayDetached"),
        core_waiter.NormalizedOperationMethod(ec2.describe_vpn_gateways),
    ),
    ("EC2", "vpn_gateway_attached"): lambda ec2: core_waiter.Waiter(
        "vpn_gateway_attached",
        ec2_model("VpnGatewayAttached"),
        core_waiter.NormalizedOperationMethod(ec2.describe_vpn_gateways),
    ),
    ("EC2", "nat_gateway_deleted"): lambda ec2: core_waiter.Waiter(
        "nat_gateway_deleted",
        ec2_model("NatGatewayDeleted"),
        core_waiter.NormalizedOperationMethod(ec2.describe_nat_gateways),
    ),
    ("EC2", "nat_gateway_available"): lambda ec2: core_waiter.Waiter(
        "nat_gateway_available",
        ec2_model("NatGatewayAvailable"),
        core_waiter.NormalizedOperationMethod(ec2.describe_nat_gateways),
    ),
    ("EC2", "volume_available"): lambda ec2: core_waiter.Waiter(
        "volume_available",
        ec2_model("VolumeAvailable"),
        core_waiter.NormalizedOperationMethod(ec2.describe_volumes),
    ),
    ("EC2", "volume_deleted"): lambda ec2: core_waiter.Waiter(
        "volume_deleted",
        ec2_model("VolumeDeleted"),
        core_waiter.NormalizedOperationMethod(ec2.describe_volumes),
    ),
    ("EC2", "volume_in_use"): lambda ec2: core_waiter.Waiter(
        "volume_in_use",
        ec2_model("VolumeInUse"),
        core_waiter.NormalizedOperationMethod(ec2.describe_volumes),
    ),
    ("WAF", "change_token_in_sync"): lambda waf: core_waiter.Waiter(
        "change_token_in_sync",
        waf_model("ChangeTokenInSync"),
        core_waiter.NormalizedOperationMethod(waf.get_change_token_status),
    ),
    ("WAFRegional", "change_token_in_sync"): lambda waf: core_waiter.Waiter(
        "change_token_in_sync",
        waf_model("ChangeTokenInSync"),
        core_waiter.NormalizedOperationMethod(waf.get_change_token_status),
    ),
    ("EKS", "cluster_active"): lambda eks: core_waiter.Waiter(
        "cluster_active", eks_model("ClusterActive"), core_waiter.NormalizedOperationMethod(eks.describe_cluster)
    ),
    ("EKS", "cluster_deleted"): lambda eks: core_waiter.Waiter(
        "cluster_deleted", eks_model("ClusterDeleted"), core_waiter.NormalizedOperationMethod(eks.describe_cluster)
    ),
    ("EKS", "fargate_profile_active"): lambda eks: core_waiter.Waiter(
        "fargate_profile_active",
        eks_model("FargateProfileActive"),
        core_waiter.NormalizedOperationMethod(eks.describe_fargate_profile),
    ),
    ("EKS", "fargate_profile_deleted"): lambda eks: core_waiter.Waiter(
        "fargate_profile_deleted",
        eks_model("FargateProfileDeleted"),
        core_waiter.NormalizedOperationMethod(eks.describe_fargate_profile),
    ),
    ("EKS", "nodegroup_active"): lambda eks: core_waiter.Waiter(
        "nodegroup_active", eks_model("NodegroupActive"), core_waiter.NormalizedOperationMethod(eks.describe_nodegroup)
    ),
    ("EKS", "nodegroup_deleted"): lambda eks: core_waiter.Waiter(
        "nodegroup_deleted",
        eks_model("NodegroupDeleted"),
        core_waiter.NormalizedOperationMethod(eks.describe_nodegroup),
    ),
    ("ElasticLoadBalancing", "any_instance_in_service"): lambda elb: core_waiter.Waiter(
        "any_instance_in_service",
        elb_model("AnyInstanceInService"),
        core_waiter.NormalizedOperationMethod(elb.describe_instance_health),
    ),
    ("ElasticLoadBalancing", "instance_deregistered"): lambda elb: core_waiter.Waiter(
        "instance_deregistered",
        elb_model("InstanceDeregistered"),
        core_waiter.NormalizedOperationMethod(elb.describe_instance_health),
    ),
    ("ElasticLoadBalancing", "instance_in_service"): lambda elb: core_waiter.Waiter(
        "load_balancer_created",
        elb_model("InstanceInService"),
        core_waiter.NormalizedOperationMethod(elb.describe_instance_health),
    ),
    ("ElasticLoadBalancing", "load_balancer_created"): lambda elb: core_waiter.Waiter(
        "load_balancer_created",
        elb_model("LoadBalancerCreated"),
        core_waiter.NormalizedOperationMethod(elb.describe_load_balancers),
    ),
    ("ElasticLoadBalancing", "load_balancer_deleted"): lambda elb: core_waiter.Waiter(
        "load_balancer_deleted",
        elb_model("LoadBalancerDeleted"),
        core_waiter.NormalizedOperationMethod(elb.describe_load_balancers),
    ),
    ("ElasticLoadBalancingv2", "load_balancer_available"): lambda elbv2: core_waiter.Waiter(
        "load_balancer_available",
        elbv2_model("LoadBalancerAvailable"),
        core_waiter.NormalizedOperationMethod(elbv2.describe_load_balancers),
    ),
    ("ElasticLoadBalancingv2", "load_balancer_ip_address_type_ipv4"): lambda elbv2: core_waiter.Waiter(
        "load_balancer_ip_address_type_ipv4",
        elbv2_model("LoadBalancerIpAddressTypeIpv4"),
        core_waiter.NormalizedOperationMethod(elbv2.describe_load_balancers),
    ),
    ("ElasticLoadBalancingv2", "load_balancer_ip_address_type_dualstack"): lambda elbv2: core_waiter.Waiter(
        "load_balancers_ip_address_type_dualstack",
        elbv2_model("LoadBalancerIpAddressTypeDualStack"),
        core_waiter.NormalizedOperationMethod(elbv2.describe_load_balancers),
    ),
    ("ElasticLoadBalancingv2", "load_balancers_deleted"): lambda elbv2: core_waiter.Waiter(
        "load_balancers_deleted",
        elbv2_model("LoadBalancersDeleted"),
        core_waiter.NormalizedOperationMethod(elbv2.describe_load_balancers),
    ),
    ("RDS", "db_cluster_promoting"): lambda rds: core_waiter.Waiter(
        "db_cluster_promoting",
        rds_model("DBClusterPromoting"),
        core_waiter.NormalizedOperationMethod(rds.describe_db_clusters),
    ),
    ("RDS", "db_instance_stopped"): lambda rds: core_waiter.Waiter(
        "db_instance_stopped",
        rds_model("DBInstanceStopped"),
        core_waiter.NormalizedOperationMethod(rds.describe_db_instances),
    ),
    ("RDS", "cluster_available"): lambda rds: core_waiter.Waiter(
        "cluster_available",
        rds_model("DBClusterAvailable"),
        core_waiter.NormalizedOperationMethod(rds.describe_db_clusters),
    ),
    ("RDS", "cluster_deleted"): lambda rds: core_waiter.Waiter(
        "cluster_deleted",
        rds_model("DBClusterDeleted"),
        core_waiter.NormalizedOperationMethod(rds.describe_db_clusters),
    ),
    ("RDS", "read_replica_promoted"): lambda rds: core_waiter.Waiter(
        "read_replica_promoted",
        rds_model("ReadReplicaPromoted"),
        core_waiter.NormalizedOperationMethod(rds.describe_db_instances),
    ),
    ("RDS", "role_associated"): lambda rds: core_waiter.Waiter(
        "role_associated", rds_model("RoleAssociated"), core_waiter.NormalizedOperationMethod(rds.describe_db_instances)
    ),
    ("RDS", "role_disassociated"): lambda rds: core_waiter.Waiter(
        "role_disassociated",
        rds_model("RoleDisassociated"),
        core_waiter.NormalizedOperationMethod(rds.describe_db_instances),
    ),
    ("Route53", "resource_record_sets_changed"): lambda route53: core_waiter.Waiter(
        "resource_record_sets_changed",
        route53_model("ResourceRecordSetsChanged"),
        core_waiter.NormalizedOperationMethod(route53.get_change),
    ),
}


def get_waiter(client, waiter_name):
    if isinstance(client, RetryingBotoClientWrapper):
        return get_waiter(client.client, waiter_name)
    try:
        return waiters_by_name[(client.__class__.__name__, waiter_name)](client)
    except KeyError:
        available_waiters = ", ".join(repr(k) for k in waiters_by_name.keys())
        raise NotImplementedError(
            f"Waiter {waiter_name} could not be found for client {type(client)}. Available waiters: {available_waiters}"
        )


def wait_for_resource_state(
    client,
    module: AnsibleAWSModule,
    waiter_name: str,
    delay: Optional[int] = None,
    max_attempts: Optional[int] = None,
    **params,
) -> None:
    waiter_params = copy.deepcopy(params)
    if delay is not None or max_attempts is not None:
        waiter_params["WaiterConfig"] = {}
        if delay is not None:
            waiter_params["WaiterConfig"].update({"Delay": delay})
        if max_attempts is not None:
            waiter_params["WaiterConfig"].update({"MaxAttempts": max_attempts})
    try:
        get_waiter(client=client, waiter_name=waiter_name).wait(**waiter_params)
    except exception_waiter.WaiterError as e:
        module.fail_json_aws(e, f"Timeout waiting for '{waiter_name}'")
    except (exception_waiter.ClientError, exception_waiter.BotoCoreError) as e:
        module.fail_json_aws(e, f"An exception happened while trying to wait for '{waiter_name}'")



# Copyright (c) 2016 Michael Gruener <michael.gruener@chaosmoon.net>
# Copyright (c) 2021 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import time
import typing as t
from collections.abc import Callable

from ansible_collections.community.crypto.plugins.module_utils._acme.challenges import (
    Authorization,
    normalize_combined_identifier,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ACMEProtocolException,
    ModuleFailException,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.utils import (
    nopad_b64,
)

if t.TYPE_CHECKING:
    from ansible_collections.community.crypto.plugins.module_utils._acme.acme import (  # pragma: no cover
        ACMEClient,
    )


_Order = t.TypeVar("_Order", bound="Order")


class Order:
    def __init__(self, *, url: str) -> None:
        self.url = url

        self.data: dict[str, t.Any] | None = None

        self.status: str | None = None
        self.identifiers: list[tuple[str, str]] = []
        self.replaces_cert_id: str | None = None
        self.finalize_uri: str | None = None
        self.certificate_uri: str | None = None
        self.authorization_uris: list[str] = []
        self.authorizations: dict[str, Authorization] = {}

    def _setup(self, *, client: ACMEClient, data: dict[str, t.Any]) -> None:
        self.data = data

        self.status = data["status"]
        self.identifiers = []
        for identifier in data["identifiers"]:
            self.identifiers.append((identifier["type"], identifier["value"]))
        self.replaces_cert_id = data.get("replaces")
        self.finalize_uri = data.get("finalize")
        self.certificate_uri = data.get("certificate")
        self.authorization_uris = data["authorizations"]
        self.authorizations = {}

    @classmethod
    def from_json(
        cls: type[_Order], *, client: ACMEClient, data: dict[str, t.Any], url: str
    ) -> _Order:
        result = cls(url=url)
        result._setup(client=client, data=data)
        return result

    @classmethod
    def from_url(cls: type[_Order], *, client: ACMEClient, url: str) -> _Order:
        result = cls(url=url)
        result.refresh(client=client)
        return result

    @classmethod
    def create(
        cls: type[_Order],
        *,
        client: ACMEClient,
        identifiers: list[tuple[str, str]],
        replaces_cert_id: str | None = None,
        profile: str | None = None,
    ) -> _Order:
        """
        Start a new certificate order (ACME v2 protocol).
        https://tools.ietf.org/html/rfc8555#section-7.4
        """
        acme_identifiers = []
        for identifier_type, identifier in identifiers:
            acme_identifiers.append(
                {
                    "type": identifier_type,
                    "value": identifier,
                }
            )
        new_order: dict[str, t.Any] = {"identifiers": acme_identifiers}
        if replaces_cert_id is not None:
            new_order["replaces"] = replaces_cert_id
        if profile is not None:
            new_order["profile"] = profile
        result, info = client.send_signed_request(
            client.directory["newOrder"],
            new_order,
            error_msg="Failed to start new order",
            expected_status_codes=[201],
        )
        if not isinstance(result, dict):
            raise ACMEProtocolException(
                module=client.module,
                msg="Unexpected new order response",
                content_json=result,
            )
        return cls.from_json(client=client, data=result, url=info["location"])

    @classmethod
    def create_with_error_handling(
        cls: type[_Order],
        *,
        client: ACMEClient,
        identifiers: list[tuple[str, str]],
        error_strategy: t.Literal[
            "auto", "fail", "always", "retry_without_replaces_cert_id"
        ] = "auto",
        error_max_retries: int = 3,
        replaces_cert_id: str | None = None,
        profile: str | None = None,
        message_callback: Callable[[str], None] | None = None,
    ) -> _Order:
        """
        error_strategy can be one of the following strings:

        * ``fail``: simply fail. (Same behavior as ``Order.create()``.)
        * ``retry_without_replaces_cert_id``: if ``replaces_cert_id`` is not ``None``, set it to ``None`` and retry.
          The only exception is an error of type ``urn:ietf:params:acme:error:alreadyReplaced``, that indicates that
          the certificate was already replaced.
        * ``auto``: try to be clever. Right now this is identical to ``retry_without_replaces_cert_id``, but that can
          change at any time in the future.
        * ``always``: always retry until ``error_max_retries`` has been reached.
        """
        tries = 0
        while True:
            tries += 1
            try:
                return cls.create(
                    client=client,
                    identifiers=identifiers,
                    replaces_cert_id=replaces_cert_id,
                    profile=profile,
                )
            except ACMEProtocolException as exc:
                if tries <= error_max_retries + 1 and error_strategy != "fail":
                    if error_strategy == "always":
                        continue

                    if (
                        error_strategy in ("auto", "retry_without_replaces_cert_id")
                        and replaces_cert_id is not None
                        and not (
                            exc.error_code == 409
                            and exc.error_type
                            == "urn:ietf:params:acme:error:alreadyReplaced"
                        )
                    ):
                        if message_callback:
                            message_callback(
                                f"Stop passing `replaces={replaces_cert_id}` due to error {exc.error_code} {exc.error_type} when creating ACME order"
                            )
                        replaces_cert_id = None
                        continue

                raise

    def refresh(self, *, client: ACMEClient) -> bool:
        result, info = client.get_request(self.url)
        if not isinstance(result, dict):
            raise ACMEProtocolException(
                module=client.module,
                msg="Unexpected authorization data",
                info=info,
                content_json=result,
            )
        changed = self.data != result
        self._setup(client=client, data=result)
        return changed

    def load_authorizations(self, *, client: ACMEClient) -> None:
        for auth_uri in self.authorization_uris:
            authz = Authorization.from_url(client=client, url=auth_uri)
            self.authorizations[
                normalize_combined_identifier(authz.combined_identifier)
            ] = authz

    def wait_for_finalization(self, *, client: ACMEClient) -> None:
        while True:
            self.refresh(client=client)
            if self.status in ["valid", "invalid", "pending", "ready"]:
                break
            time.sleep(2)

        if self.status != "valid":
            raise ACMEProtocolException(
                module=client.module,
                msg=f'Failed to wait for order to complete; got status "{self.status}"',
                content_json=self.data,
            )

    def finalize(
        self, *, client: ACMEClient, csr_der: bytes, wait: bool = True
    ) -> None:
        """
        Create a new certificate based on the csr.
        Return the certificate object as dict
        https://tools.ietf.org/html/rfc8555#section-7.4
        """
        if self.finalize_uri is None:
            raise ModuleFailException("finalize_uri must be set")
        new_cert = {
            "csr": nopad_b64(csr_der),
        }
        result, info = client.send_signed_request(
            self.finalize_uri,
            new_cert,
            error_msg="Failed to finalizing order",
            expected_status_codes=[200],
        )
        # It is not clear from the RFC whether the finalize call returns the order object or not.
        # Instead of using the result, we call self.refresh(client) below.

        if wait:
            self.wait_for_finalization(client=client)
        else:
            self.refresh(client=client)
            if self.status not in ["procesing", "valid", "invalid"]:
                raise ACMEProtocolException(
                    module=client.module,
                    msg=f'Failed to finalize order; got status "{self.status}"',
                    info=info,
                    content_json=result,
                )


__all__ = ("Order",)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Akini Ross (@akinross)
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["preview"],
    "supported_by": "community",
}

DOCUMENTATION = r"""
---
module: aci_interface_policy_port_channel_member
version_added: "2.12.0"
short_description: Manage Port Channel Member interface policies (lacp:IfPol)
description:
- Manage Port Channel Member interface policy configuration on Cisco ACI fabrics.
options:
  name:
    description:
    - The name of the Port Channel Member interface policy.
    type: str
    aliases: [ port_channel_member_interface_policy ]
  description:
    description:
    - The description of the Port Channel Member interface policy.
    type: str
  priority:
    description:
    - The priority of the Port Channel Member interface policy.
    - The APIC defaults to C(32768) when not provided.
    - Accepted values range between C(1) and C(65535).
    type: int
  transmit_rate:
    description:
    - The transmit rate of the Port Channel Member interface policy.
    - The APIC defaults to C(normal) when not provided.
    type: str
    choices: [ normal, fast ]
  name_alias:
    description:
    - The alias for the current object.
    - This relates to the nameAlias field in ACI.
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
- cisco.aci.owner

seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(lacp:IfPol).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Akini Ross (@akinross)
"""

EXAMPLES = r"""
- name: Add a new Port Channel Member interface policy
  cisco.aci.aci_interface_policy_port_channel_member:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: ansible_port_channel_member_policy
    description: Ansible Port Channel Member interface policy
    priority: 32700
    transmit_rate: fast
    state: present
  delegate_to: localhost

- name: Query a Port Channel Member interface policy
  cisco.aci.aci_interface_policy_port_channel_member:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: ansible_port_channel_member_policy
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all Port Channel Member interface policies
  cisco.aci.aci_interface_policy_port_channel_member:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result

- name: Remove a Port Channel Member interface policy
  cisco.aci.aci_interface_policy_port_channel_member:
    host: apic
    username: admin
    password: SomeSecretPassword
    name: my_ntp_policy
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
   current:
     description: The existing configuration from the APIC after the module has finished
     returned: success
     type: list
     sample:
       [
           {
               "fvTenant": {
                   "attributes": {
                       "descr": "Production environment",
                       "dn": "uni/tn-production",
                       "name": "production",
                       "nameAlias": "",
                       "ownerKey": "",
                       "ownerTag": ""
                   }
               }
           }
       ]
   error:
     description: The error information as returned from the APIC
     returned: failure
     type: dict
     sample:
       {
           "code": "122",
           "text": "unknown managed object class foo"
       }
   raw:
     description: The raw output returned by the APIC REST API (xml or json)
     returned: parse error
     type: str
     sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class "/></imdata>'
   sent:
     description: The actual/minimal configuration pushed to the APIC
     returned: info
     type: list
     sample:
       {
           "fvTenant": {
               "attributes": {
                   "descr": "Production environment"
               }
           }
       }
   previous:
     description: The original configuration from the APIC before the module has started
     returned: info
     type: list
     sample:
       [
           {
               "fvTenant": {
                   "attributes": {
                       "descr": "Production",
                       "dn": "uni/tn-production",
                       "name": "production",
                       "nameAlias": "",
                       "ownerKey": "",
                       "ownerTag": ""
                   }
               }
           }
       ]
   proposed:
     description: The assembled configuration from the user-provided parameters
     returned: info
     type: dict
     sample:
       {
           "fvTenant": {
               "attributes": {
                   "descr": "Production environment",
                   "name": "production"
               }
           }
       }
   filter_string:
     description: The filter string used for the request
     returned: failure or debug
     type: str
     sample: ?rsp-prop-include=config-only
   method:
     description: The HTTP method used for the request to the APIC
     returned: failure or debug
     type: str
     sample: POST
   response:
     description: The HTTP response from the APIC
     returned: failure or debug
     type: str
     sample: OK (30 bytes)
   status:
     description: The HTTP status from the APIC
     returned: failure or debug
     type: int
     sample: 200
   url:
     description: The HTTP url used for the request to the APIC
     returned: failure or debug
     type: str
     sample: https://10.11.12.13/api/mo/uni/tn-production.json
   """

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec, aci_owner_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(aci_owner_spec())
    argument_spec.update(
        name=dict(type="str", aliases=["port_channel_member_interface_policy"]),
        description=dict(type="str"),
        priority=dict(type="int"),
        transmit_rate=dict(type="str", choices=["normal", "fast"]),
        name_alias=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["name"]],
            ["state", "present", ["name"]],
        ],
    )

    name = module.params.get("name")
    description = module.params.get("description")
    priority = module.params.get("priority")
    transmit_rate = module.params.get("transmit_rate")
    name_alias = module.params.get("name_alias")
    state = module.params.get("state")

    aci = ACIModule(module)
    aci.construct_url(
        root_class=dict(
            aci_class="lacpIfPol",
            aci_rn="infra/lacpifp-{0}".format(name),
            module_object=name,
            target_filter={"name": name},
        ),
    )

    aci.get_existing()

    if state == "present":
        aci.payload(
            aci_class="lacpIfPol",
            class_config=dict(
                name=name,
                descr=description,
                prio=priority,
                txRate=transmit_rate,
                nameAlias=name_alias,
            ),
        )

        aci.get_diff(aci_class="lacpIfPol")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2019, Kaarle Ritvanen <kaarle.ritvanen@datakunkku.fi>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: lbu

short_description: Local Backup Utility for Alpine Linux

version_added: '0.2.0'

description:
  - Manage Local Backup Utility of Alpine Linux in run-from-RAM mode.
extends_documentation_fragment:
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  commit:
    description:
      - Control whether to commit changed files.
    type: bool
  exclude:
    description:
      - List of paths to exclude.
    type: list
    elements: str
  include:
    description:
      - List of paths to include.
    type: list
    elements: str

author:
  - Kaarle Ritvanen (@kunkku)
"""

EXAMPLES = r"""
# Commit changed files (if any)
- name: Commit
  community.general.lbu:
    commit: true

# Exclude path and commit
- name: Exclude directory
  community.general.lbu:
    commit: true
    exclude:
      - /etc/opt

# Include paths without committing
- name: Include file and directory
  community.general.lbu:
    include:
      - /root/.ssh/authorized_keys
      - /var/lib/misc
"""

RETURN = r"""
msg:
  description: Error message.
  type: str
  returned: on failure
"""

import os.path

from ansible.module_utils.basic import AnsibleModule


def run_module():
    module = AnsibleModule(
        argument_spec={
            "commit": {"type": "bool"},
            "exclude": {"type": "list", "elements": "str"},
            "include": {"type": "list", "elements": "str"},
        },
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    changed = False

    def run_lbu(*args):
        code, stdout, stderr = module.run_command([module.get_bin_path("lbu", required=True)] + list(args))
        if code:
            module.fail_json(changed=changed, msg=stderr)
        return stdout

    update = False
    commit = False

    for param in ("include", "exclude"):
        if module.params[param]:
            paths = run_lbu(param, "-l").split("\n")
            for path in module.params[param]:
                if os.path.normpath(f"/{path}")[1:] not in paths:
                    update = True

    if module.params["commit"]:
        commit = update or run_lbu("status") > ""

    if module.check_mode:
        module.exit_json(changed=update or commit)

    if update:
        for param in ("include", "exclude"):
            if module.params[param]:
                run_lbu(param, *module.params[param])
                changed = True

    if commit:
        run_lbu("commit")
        changed = True

    module.exit_json(changed=changed)


def main():
    run_module()


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright 2017 Red Hat | Ansible, Alex Grönholm <alex.gronholm@nextday.fi>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_volume_info
short_description: Retrieve facts about Docker volumes
description:
  - Performs largely the same function as the C(docker volume inspect) CLI subcommand.
extends_documentation_fragment:
  - community.docker._docker.api_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker
  - community.docker._attributes.info_module
  - community.docker._attributes.idempotent_not_modify_state

options:
  name:
    description:
      - Name of the volume to inspect.
    type: str
    required: true
    aliases:
      - volume_name

author:
  - Felix Fontein (@felixfontein)

requirements:
  - "Docker API >= 1.25"
"""

EXAMPLES = r"""
---
- name: Get infos on volume
  community.docker.docker_volume_info:
    name: mydata
  register: result

- name: Does volume exist?
  ansible.builtin.debug:
    msg: "The volume {{ 'exists' if result.exists else 'does not exist' }}"

- name: Print information about volume
  ansible.builtin.debug:
    var: result.volume
  when: result.exists
"""

RETURN = r"""
exists:
  description:
    - Returns whether the volume exists.
  type: bool
  returned: always
  sample: true
volume:
  description:
    - Volume inspection results for the affected volume.
    - Will be V(none) if volume does not exist.
  returned: success
  type: dict
  sample: '{ "CreatedAt": "2018-12-09T17:43:44+01:00", "Driver": "local", "Labels": null, "Mountpoint": "/var/lib/docker/volumes/ansible-test-bd3f6172/_data",
    "Name": "ansible-test-bd3f6172", "Options": {}, "Scope": "local" }'
"""

import traceback
import typing as t

from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    DockerException,
    NotFound,
)
from ansible_collections.community.docker.plugins.module_utils._common_api import (
    AnsibleDockerClient,
    RequestException,
)


def get_existing_volume(
    client: AnsibleDockerClient, volume_name: str
) -> dict[str, t.Any] | None:
    try:
        return client.get_json("/volumes/{0}", volume_name)
    except NotFound:
        return None
    except Exception as exc:  # pylint: disable=broad-exception-caught
        client.fail(f"Error inspecting volume: {exc}")


def main() -> None:
    argument_spec = {
        "name": {"type": "str", "required": True, "aliases": ["volume_name"]},
    }

    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    try:
        volume = get_existing_volume(client, client.module.params["name"])

        client.module.exit_json(
            changed=False,
            exists=bool(volume),
            volume=volume,
        )
    except DockerException as e:
        client.fail(
            f"An unexpected Docker error occurred: {e}",
            exception=traceback.format_exc(),
        )
    except RequestException as e:
        client.fail(
            f"An unexpected requests error occurred when trying to talk to the Docker daemon: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2022 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# cli_rm_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the module docstring and re-run
# cli_rm_builder.
#
#############################################

"""
The arg spec for the iosxr_snmp_server module
"""


class Snmp_serverArgs(object):  # pylint: disable=R0903
    """The arg spec for the iosxr_snmp_server module"""

    argument_spec = {
        "config": {
            "type": "dict",
            "options": {
                "chassis_id": {"type": "str"},
                "communities": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "name": {"type": "str"},
                        "acl_v4": {"type": "str"},
                        "acl_v6": {"type": "str"},
                        "ro": {"type": "bool"},
                        "rw": {"type": "bool"},
                        "sdrowner": {"type": "bool"},
                        "systemowner": {"type": "bool"},
                        "v4_acl": {"type": "str"},
                    },
                },
                "community_maps": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "name": {"type": "str"},
                        "context": {"type": "str"},
                        "security_name": {"type": "str"},
                        "target_list": {"type": "str"},
                    },
                },
                "correlator": {
                    "type": "dict",
                    "options": {
                        "buffer_size": {"type": "int"},
                        "rules": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "rule_name": {"type": "str"},
                                "timeout": {"type": "int"},
                            },
                        },
                        "rule_sets": {
                            "type": "list",
                            "elements": "dict",
                            "options": {"name": {"type": "str"}},
                        },
                    },
                },
                "contact": {"type": "str"},
                "context": {"type": "list", "elements": "str"},
                "drop": {
                    "type": "dict",
                    "options": {
                        "unknown_user": {"type": "bool"},
                        "report_IPv4": {"type": "str"},
                        "report_IPv6": {"type": "str"},
                    },
                },
                "engineid": {
                    "type": "dict",
                    "options": {"local": {"type": "str"}},
                },
                "groups": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "group": {"type": "str"},
                        "version": {
                            "type": "str",
                            "choices": ["v1", "v3", "v2c"],
                        },
                        "context": {"type": "str"},
                        "notify": {"type": "str"},
                        "read": {"type": "str"},
                        "write": {"type": "str"},
                        "acl_v4": {"type": "str", "aliases": ["Ipv4_acl"]},
                        "acl_v6": {"type": "str", "aliases": ["Ipv6_acl"]},
                        "v4_acl": {"type": "str"},
                    },
                },
                "hosts": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "host": {"type": "str"},
                        "community": {"type": "str"},
                        "udp_port": {"type": "int"},
                        "informs": {"type": "bool"},
                        "traps": {"type": "bool"},
                        "version": {
                            "type": "str",
                            "choices": ["1", "2c", "3"],
                        },
                    },
                },
                "ifindex": {"type": "bool"},
                "ifmib": {
                    "type": "dict",
                    "options": {
                        "ifalias_long": {"type": "bool"},
                        "internal_cache_max_duration": {"type": "int"},
                        "ipsubscriber": {"type": "bool"},
                        "stats": {"type": "bool"},
                    },
                },
                "inform": {
                    "options": {
                        "pending": {"type": "int"},
                        "retries": {"type": "int"},
                        "timeout": {"type": "int"},
                    },
                    "type": "dict",
                },
                "interfaces": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "name": {"type": "str"},
                        "notification_linkupdown_disable": {"type": "bool"},
                        "index_persistent": {"type": "bool"},
                    },
                },
                "ipv4": {
                    "type": "dict",
                    "options": {
                        "dscp": {"type": "str"},
                        "precedence": {"type": "str"},
                    },
                },
                "ipv6": {
                    "type": "dict",
                    "options": {
                        "dscp": {"type": "str"},
                        "precedence": {"type": "str"},
                    },
                },
                "location": {"type": "str"},
                "logging_threshold_oid_processing": {"type": "int"},
                "logging_threshold_pdu_processing": {"type": "int"},
                "mib_bulkstat_max_procmem_size": {"type": "int"},
                "mib_object_lists": {"type": "list", "elements": "str"},
                "mib_schema": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "name": {"type": "str"},
                        "object_list": {"type": "str"},
                        "poll_interval": {"type": "int"},
                    },
                },
                "mib_bulkstat_transfer_ids": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "name": {"type": "str"},
                        "buffer_size": {"type": "int"},
                        "enable": {"type": "bool"},
                        "format_schemaASCI": {"type": "bool"},
                        "retain": {"type": "int"},
                        "retry": {"type": "int"},
                        "schema": {"type": "str"},
                        "transfer_interval": {"type": "int"},
                    },
                },
                "mroutemib_send_all_vrf": {"type": "bool"},
                "notification_log_mib": {
                    "type": "dict",
                    "options": {
                        "GlobalSize": {"type": "int"},
                        "default": {"type": "bool"},
                        "disable": {"type": "bool"},
                        "size": {"type": "int"},
                    },
                },
                "oid_poll_stats": {"type": "bool"},
                "overload_control": {
                    "type": "dict",
                    "options": {
                        "overload_drop_time": {"type": "int"},
                        "overload_throttle_rate": {"type": "int"},
                    },
                },
                "packetsize": {"type": "int"},
                "queue_length": {"type": "int"},
                "targets": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "name": {"type": "str"},
                        "host": {"type": "str"},
                        "vrf": {"type": "str"},
                    },
                },
                "throttle_time": {"type": "int"},
                "timeouts": {
                    "type": "dict",
                    "options": {
                        "duplicate": {"type": "int"},
                        "inQdrop": {"type": "int"},
                        "pdu_stats": {"type": "int"},
                        "subagent": {"type": "int"},
                        "threshold": {"type": "int"},
                    },
                },
                "trap": {
                    "type": "dict",
                    "options": {
                        "authentication_vrf_disable": {"type": "bool"},
                        "link_ietf": {"type": "bool"},
                        "throttle_time": {"type": "int"},
                    },
                },
                "trap_source": {"type": "str"},
                "trap_timeout": {"type": "int"},
                "traps": {
                    "type": "dict",
                    "options": {
                        "addrpool": {
                            "type": "dict",
                            "options": {
                                "low": {"type": "bool"},
                                "high": {"type": "bool"},
                            },
                        },
                        "bfd": {"type": "bool"},
                        "bgp": {
                            "type": "dict",
                            "options": {
                                "cbgp2": {"type": "bool"},
                                "updown": {"type": "bool"},
                            },
                        },
                        "bulkstat_collection": {"type": "bool"},
                        "bulkstat_transfer": {"type": "bool"},
                        "bridgemib": {"type": "bool"},
                        "copy_complete": {"type": "bool"},
                        "cisco_entity_ext": {"type": "bool"},
                        "config": {"type": "bool"},
                        "diameter": {
                            "type": "dict",
                            "options": {
                                "peerdown": {"type": "bool"},
                                "peerup": {"type": "bool"},
                                "permanentfail": {"type": "bool"},
                                "protocolerror": {"type": "bool"},
                                "transientfail": {"type": "bool"},
                            },
                        },
                        "entity": {"type": "bool"},
                        "entity_redundancy": {
                            "type": "dict",
                            "options": {
                                "all": {"type": "bool"},
                                "status": {"type": "bool"},
                                "switchover": {"type": "bool"},
                            },
                        },
                        "entity_state": {
                            "type": "dict",
                            "options": {
                                "operstatus": {"type": "bool"},
                                "switchover": {"type": "bool"},
                            },
                        },
                        "flash": {
                            "type": "dict",
                            "options": {
                                "insertion": {"type": "bool"},
                                "removal": {"type": "bool"},
                            },
                        },
                        "fru_ctrl": {"type": "bool"},
                        "hsrp": {"type": "bool"},
                        "ipsla": {"type": "bool"},
                        "ipsec": {
                            "type": "dict",
                            "options": {
                                "start": {"type": "bool"},
                                "stop": {"type": "bool"},
                            },
                        },
                        "isakmp": {
                            "type": "dict",
                            "options": {
                                "start": {"type": "bool"},
                                "stop": {"type": "bool"},
                            },
                        },
                        "isis": {
                            "type": "dict",
                            "options": {
                                "adjacency_change": {"type": "bool"},
                                "all": {"type": "bool"},
                                "area_mismatch": {"type": "bool"},
                                "attempt_to_exceed_max_sequence": {
                                    "type": "bool",
                                },
                                "authentication_failure": {"type": "bool"},
                                "authentication_type_failure": {
                                    "type": "bool",
                                },
                                "corrupted_lsp_detected": {"type": "bool"},
                                "database_overload": {"type": "bool"},
                                "id_len_mismatch": {"type": "bool"},
                                "lsp_error_detected": {"type": "bool"},
                                "lsp_too_large_to_propagate": {"type": "bool"},
                                "manual_address_drops": {"type": "bool"},
                                "max_area_addresses_mismatch": {
                                    "type": "bool",
                                },
                                "orig_lsp_buff_size_mismatch": {
                                    "type": "bool",
                                },
                                "version_skew": {"type": "bool"},
                                "own_lsp_purge": {"type": "bool"},
                                "rejected_adjacency": {"type": "bool"},
                                "protocols_supported_mismatch": {
                                    "type": "bool",
                                },
                                "sequence_number_skip": {"type": "bool"},
                            },
                        },
                        "l2tun": {
                            "type": "dict",
                            "options": {
                                "pseudowire_status": {"type": "bool"},
                                "sessions": {"type": "bool"},
                                "tunnel_down": {"type": "bool"},
                                "tunnel_up": {"type": "bool"},
                            },
                        },
                        "l2vpn": {
                            "type": "dict",
                            "options": {
                                "all": {"type": "bool"},
                                "cisco": {"type": "bool"},
                                "vc_down": {"type": "bool"},
                                "vc_up": {"type": "bool"},
                            },
                        },
                        "msdp_peer_state_change": {"type": "bool"},
                        "ntp": {"type": "bool"},
                        "ospf": {
                            "type": "dict",
                            "options": {
                                "errors": {
                                    "type": "dict",
                                    "options": {
                                        "bad_packet": {"type": "bool"},
                                        "authentication_failure": {
                                            "type": "bool",
                                        },
                                        "config_error": {"type": "bool"},
                                        "virt_bad_packet": {"type": "bool"},
                                        "virt_authentication_failure": {
                                            "type": "bool",
                                        },
                                        "virt_config_error": {"type": "bool"},
                                    },
                                },
                                "lsa": {
                                    "type": "dict",
                                    "options": {
                                        "lsa_maxage": {"type": "bool"},
                                        "lsa_originate": {"type": "bool"},
                                    },
                                },
                                "retransmit": {
                                    "type": "dict",
                                    "options": {
                                        "packets": {"type": "bool"},
                                        "virt_packets": {"type": "bool"},
                                    },
                                },
                                "state_change": {
                                    "type": "dict",
                                    "options": {
                                        "if_state_change": {"type": "bool"},
                                        "neighbor_state_change": {
                                            "type": "bool",
                                        },
                                        "virtif_state_change": {
                                            "type": "bool",
                                        },
                                        "virtneighbor_state_change": {
                                            "type": "bool",
                                        },
                                    },
                                },
                            },
                        },
                        "ospfv3": {
                            "type": "dict",
                            "options": {
                                "errors": {
                                    "type": "dict",
                                    "options": {
                                        "bad_packet": {"type": "bool"},
                                        "config_error": {"type": "bool"},
                                        "virt_bad_packet": {"type": "bool"},
                                        "virt_config_error": {"type": "bool"},
                                    },
                                },
                                "state_change": {
                                    "type": "dict",
                                    "options": {
                                        "if_state_change": {"type": "bool"},
                                        "neighbor_state_change": {
                                            "type": "bool",
                                        },
                                        "virtif_state_change": {
                                            "type": "bool",
                                        },
                                        "virtneighbor_state_change": {
                                            "type": "bool",
                                        },
                                        "nssa_state_change": {"type": "bool"},
                                        "restart_status_change": {
                                            "type": "bool",
                                        },
                                        "restart_helper_status_change": {
                                            "type": "bool",
                                        },
                                        "restart_virtual_helper_status_change": {
                                            "type": "bool",
                                        },
                                    },
                                },
                            },
                        },
                        "power": {"type": "bool"},
                        "rf": {"type": "bool"},
                        "pim": {
                            "type": "dict",
                            "options": {
                                "interface_state_change": {"type": "bool"},
                                "invalid_message_received": {"type": "bool"},
                                "neighbor_change": {"type": "bool"},
                                "rp_mapping_change": {"type": "bool"},
                            },
                        },
                        "rsvp": {
                            "type": "dict",
                            "options": {
                                "all": {"type": "bool"},
                                "lost_flow": {"type": "bool"},
                                "new_flow": {"type": "bool"},
                            },
                        },
                        "selective_vrf_download_role_change": {"type": "bool"},
                        "sensor": {"type": "bool"},
                        "snmp": {
                            "type": "dict",
                            "options": {
                                "authentication": {"type": "bool"},
                                "linkdown": {"type": "bool"},
                                "linkup": {"type": "bool"},
                                "warmstart": {"type": "bool"},
                                "coldstart": {"type": "bool"},
                            },
                        },
                        "vrrp_events": {"type": "bool"},
                        "syslog": {"type": "bool"},
                        "subscriber": {
                            "type": "dict",
                            "options": {
                                "session_agg_access_interface": {
                                    "type": "bool",
                                },
                                "session_agg_node": {"type": "bool"},
                            },
                        },
                        "system": {"type": "bool"},
                        "vpls": {
                            "type": "dict",
                            "options": {
                                "all": {"type": "bool"},
                                "full_clear": {"type": "bool"},
                                "full_raise": {"type": "bool"},
                                "status": {"type": "bool"},
                            },
                        },
                    },
                },
                "users": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "user": {"type": "str"},
                        "group": {"type": "str"},
                        "acl_v4": {"type": "str", "aliases": ["Ipv4_acl"]},
                        "acl_v6": {"type": "str", "aliases": ["Ipv6_acl"]},
                        "SDROwner": {"type": "bool"},
                        "SystemOwner": {"type": "bool"},
                        "v4_acl": {"type": "str"},
                        "version": {
                            "type": "str",
                            "choices": ["v1", "v2c", "v3"],
                        },
                    },
                },
                "vrfs": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "vrf": {"type": "str"},
                        "context": {"type": "list", "elements": "str"},
                        "hosts": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "host": {"type": "str"},
                                "community": {"type": "str"},
                                "udp_port": {"type": "int"},
                                "informs": {"type": "bool"},
                                "traps": {"type": "bool"},
                                "version": {
                                    "type": "str",
                                    "choices": ["1", "2c", "3"],
                                },
                            },
                        },
                    },
                },
            },
        },
        "running_config": {"type": "str"},
        "state": {
            "type": "str",
            "choices": [
                "deleted",
                "merged",
                "overridden",
                "replaced",
                "gathered",
                "rendered",
                "parsed",
            ],
            "default": "merged",
        },
    }  # pylint: disable=C0301



# This code is part of the Ansible collection community.docker, but is an independent component.
# This particular file, and this file only, is based on the Docker SDK for Python (https://github.com/docker/docker-py/)
#
# Copyright (c) 2016-2022 Docker, Inc.
#
# It is licensed under the Apache 2.0 license (see LICENSES/Apache-2.0.txt in this collection)
# SPDX-License-Identifier: Apache-2.0

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import os
import typing as t

from . import errors
from .transport.ssladapter import SSLHTTPAdapter

if t.TYPE_CHECKING:
    from ansible_collections.community.docker.plugins.module_utils._api.api.client import (
        APIClient,
    )


class TLSConfig:
    """
    TLS configuration.

    Args:
        client_cert (tuple of str): Path to client cert, path to client key.
        ca_cert (str): Path to CA cert file.
        verify (bool or str): This can be ``False`` or a path to a CA cert
            file.
        assert_hostname (bool): Verify the hostname of the server.

    .. _`SSL version`:
        https://docs.python.org/3.5/library/ssl.html#ssl.PROTOCOL_TLSv1
    """

    cert: tuple[str, str] | None = None
    ca_cert: str | None = None
    verify: bool | None = None

    def __init__(
        self,
        client_cert: tuple[str, str] | None = None,
        ca_cert: str | None = None,
        verify: bool | None = None,
        assert_hostname: bool | None = None,
    ):
        # Argument compatibility/mapping with
        # https://docs.docker.com/engine/articles/https/
        # This diverges from the Docker CLI in that users can specify 'tls'
        # here, but also disable any public/default CA pool verification by
        # leaving verify=False

        self.assert_hostname = assert_hostname

        # "client_cert" must have both or neither cert/key files. In
        # either case, Alert the user when both are expected, but any are
        # missing.

        if client_cert:
            try:
                tls_cert, tls_key = client_cert
            except ValueError:
                raise errors.TLSParameterError(
                    "client_cert must be a tuple of (client certificate, key file)"
                ) from None

            if not (tls_cert and tls_key) or (
                not os.path.isfile(tls_cert) or not os.path.isfile(tls_key)
            ):
                raise errors.TLSParameterError(
                    "Path to a certificate and key files must be provided"
                    " through the client_cert param"
                )
            self.cert = (tls_cert, tls_key)

        # If verify is set, make sure the cert exists
        self.verify = verify
        self.ca_cert = ca_cert
        if self.verify and self.ca_cert and not os.path.isfile(self.ca_cert):
            raise errors.TLSParameterError(
                "Invalid CA certificate provided for `ca_cert`."
            )

    def configure_client(self, client: APIClient) -> None:
        """
        Configure a client with these TLS options.
        """

        if self.verify and self.ca_cert:
            client.verify = self.ca_cert
        else:
            client.verify = self.verify

        if self.cert:
            client.cert = self.cert

        client.mount(
            "https://",
            SSLHTTPAdapter(
                assert_hostname=self.assert_hostname,
            ),
        )



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
filter plugin file for ipaddr filters: ipwrap
"""
from __future__ import absolute_import, division, print_function

import types

from functools import partial

from ansible.errors import AnsibleFilterError

from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
    AnsibleArgSpecValidator,
)
from ansible_collections.ansible.utils.plugins.plugin_utils.base.ipaddr_utils import (
    _need_netaddr,
    ipaddr,
)


__metaclass__ = type


try:
    from jinja2.filters import pass_environment
except ImportError:
    from jinja2.filters import environmentfilter as pass_environment

try:
    import netaddr

    HAS_NETADDR = True
except ImportError:
    # in this case, we'll make the filters return error messages (see bottom)
    HAS_NETADDR = False
else:

    class mac_linux(netaddr.mac_unix):
        pass

    mac_linux.word_fmt = "%.2x"

DOCUMENTATION = """
    name: ipwrap
    author: Ashwini Mhatre (@amhatre)
    version_added: "2.5.0"
    short_description: This filter is designed to Wrap IPv6 addresses in [ ] brackets.
    description:
    - Some configuration files require IPv6 addresses to be "wrapped" in square brackets ([ ]).To accomplish that,
    - you can use the ipwrap() filter.It will wrap all IPv6 addresses and leave any other strings intact.
    options:
        value:
            description:
            - list of subnets or individual address or any other values input. Example. ['192.24.2.1', 'host.fqdn',
              '::1', '192.168.32.0/24', 'fe80::100/10', True, '', '42540766412265424405338506004571095040/64']
            type: raw
            required: True
        query:
            description:
            - You can provide a single argument to each ipwrap() filter.
            - The filter will then treat it as a query and return values modified by that query.
            type: str
            default: ''
    notes:
"""

EXAMPLES = r"""
#### examples
# Ipwrap filter plugin o Wrap IPv6 addresses in [ ] brackets.
- name: Set value as input list
  ansible.builtin.set_fact:
    value:
      - 192.24.2.1
      - host.fqdn
      - ::1
      - ''
      - 192.168.32.0/24
      - fe80::100/10
      - 42540766412265424405338506004571095040/64
      - true
- debug:
    msg: "{{ value|ansible.utils.ipwrap }}"

- name: |
        ipwrap() did not filter out non-IP address values, which is usually what you want when for example
        you are mixing IP addresses with hostnames. If you still want to filter out all non-IP address values,
        you can chain both filters together.
  debug:
    msg: "{{ value|ansible.utils.ipaddr|ansible.utils.ipwrap  }}"

# PLAY [Ipwrap filter plugin o Wrap IPv6 addresses in [ ] brackets.] ***************************************************
# TASK [Set value as input list] ***************************************************************************************
# ok: [localhost] => {"ansible_facts": {"value": ["192.24.2.1", "host.fqdn", "::1", "", "192.168.32.0/24",
# "fe80::100/10", "42540766412265424405338506004571095040/64", true]}, "changed": false}
#
# TASK [debug] ********************************************************************************************************
# ok: [localhost] => {
#     "msg": [
#         "192.24.2.1",
#         "::1",
#         "192.168.32.0/24",
#         "fe80::100/10",
#         "2001:db8:32c:faad::/64"
#     ]
# }
#
# TASK [debug] ************************************************************************************************
# ok: [localhost] => {
#     "msg": [
#         "192.24.2.1",
#         "host.fqdn",
#         "[::1]",
#         "",
#         "192.168.32.0/24",
#         "[fe80::100]/10",
#         "[2001:db8:32c:faad::]/64",
#         "True"
#     ]
# }
#
# TASK [ipwrap() did not filter out non-IP address values, which is usually what you want when for example
# you are mixing IP addresses with hostnames. If you still want to filter out all non-IP address values,
# you can chain both filters together.] ***
# ok: [localhost] => {
#     "msg": [
#         "192.24.2.1",
#         "[::1]",
#         "192.168.32.0/24",
#         "[fe80::100]/10",
#         "[2001:db8:32c:faad::]/64"
#     ]
# }
"""

RETURN = """
  data:
    type: raw
    description:
      - Returns values valid for a particular query.
"""


@pass_environment
def _ipwrap(*args, **kwargs):
    """This filter is designed to Wrap IPv6 addresses in [ ] brackets."""
    keys = ["value"]
    data = dict(zip(keys, args[1:]))
    data.update(kwargs)
    try:
        if isinstance(data["value"], str):
            pass
        elif isinstance(data["value"], list):
            pass
        elif isinstance(data["value"], bool):
            pass
        else:
            raise AnsibleFilterError(
                "Unrecognized type <{0}> for ipwrap filter <{1}>".format(
                    type(data["value"]),
                    "value",
                ),
            )

    except (TypeError, ValueError):
        raise AnsibleFilterError(
            "Unrecognized type <{0}> for ipwrap filter <{1}>".format(type(data["value"]), "value"),
        )
    aav = AnsibleArgSpecValidator(data=data, schema=DOCUMENTATION, name="ipwrap")
    valid, errors, updated_data = aav.validate()
    if not valid:
        raise AnsibleFilterError(errors)
    return ipwrap(**updated_data)


def ipwrap(value, query=""):
    try:
        if isinstance(value, (list, tuple, types.GeneratorType)):
            _ret = []
            for element in value:
                if ipaddr(element, query, version=False, alias="ipwrap"):
                    _ret.append(ipaddr(element, "wrap"))
                else:
                    _ret.append(element)

            return _ret
        else:
            _ret = ipaddr(value, query, version=False, alias="ipwrap")
            if _ret:
                return ipaddr(_ret, "wrap")
            else:
                return value

    except Exception:
        return value


class FilterModule(object):
    """IP address and network manipulation filters"""

    filter_map = {
        # IP addresses and networks
        "ipwrap": _ipwrap,
    }

    def filters(self):
        """ipwrap filter"""
        if HAS_NETADDR:
            return self.filter_map
        else:
            return dict((f, partial(_need_netaddr, f)) for f in self.filter_map)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Bruno Calogero <brunocalogero@hotmail.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_static_binding_to_epg
short_description: Bind static paths to EPGs (fv:RsPathAtt)
description:
- Bind static paths to EPGs on Cisco ACI fabrics.
options:
  tenant:
    description:
    - Name of an existing tenant.
    type: str
    aliases: [ tenant_name ]
  ap:
    description:
    - Name of an existing application network profile, that will contain the EPGs.
    type: str
    aliases: [ app_profile, app_profile_name ]
  epg:
    description:
    - The name of the end point group.
    type: str
    aliases: [ epg_name ]
  description:
    description:
    - Description for the static path to EPG binding.
    type: str
    aliases: [ descr ]
  encap_id:
    description:
    - The encapsulation ID associating the C(epg) with the interface path.
    - This acts as the secondary C(encap_id) when using micro-segmentation.
    - Accepted values are any valid encap ID for specified encap, currently ranges between C(1) and C(4096).
    type: int
    aliases: [ vlan, vlan_id ]
  primary_encap_id:
    description:
    - Determines the primary encapsulation ID associating the C(epg)
      with the interface path when using micro-segmentation.
    - Accepted values are any valid encap ID for specified encap, currently ranges between C(1) and C(4096) and C(unknown).
    - C(unknown) is the default value and using C(unknown) disables the Micro-Segmentation.
    type: str
    aliases: [ primary_vlan, primary_vlan_id ]
  deploy_immediacy:
    description:
    - The Deployment Immediacy of Static EPG on PC, VPC or Interface.
    - The APIC defaults to C(lazy) when unset during creation.
    type: str
    choices: [ immediate, lazy ]
  interface_mode:
    description:
    - Determines how layer 2 tags will be read from and added to frames.
    - Values C(802.1p) and C(native) are identical.
    - Values C(access) and C(untagged) are identical.
    - Values C(regular), C(tagged) and C(trunk) are identical.
    - The APIC defaults to C(trunk) when unset during creation.
    type: str
    choices: [ 802.1p, access, native, regular, tagged, trunk, untagged ]
    aliases: [ interface_mode_name, mode ]
  interface_type:
    description:
    - The type of interface for the static EPG deployment.
    type: str
    choices: [ fex, port_channel, switch_port, vpc, fex_port_channel, fex_vpc ]
    default: switch_port
  pod_id:
    description:
    - The pod number part of the tDn.
    - C(pod_id) is usually an integer below C(10).
    type: int
    aliases: [ pod, pod_number ]
  leafs:
    description:
    - The switch ID(s) that the C(interface) belongs to.
    - When C(interface_type) is C(switch_port), C(port_channel), or C(fex), then C(leafs) is a string of the leaf ID.
    - When C(interface_type) is C(vpc), then C(leafs) is a list with both leaf IDs.
    - The C(leafs) value is usually something like '101' or '101-102' depending on C(connection_type).
    type: list
    elements: str
    aliases: [ leaves, nodes, paths, switches ]
  interface:
    description:
    - The C(interface) string value part of the tDn.
    - Usually a policy group like C(test-IntPolGrp) or an interface of the following format C(1/7) depending on C(interface_type).
    type: str
  extpaths:
    description:
    - The C(extpaths) integer value part of the tDn.
    - C(extpaths) is only used if C(interface_type) is C(fex), C(fex_vpc) or C(fex_port_channel).
    - When C(interface_type) is C(fex_vpc), then C(extpaths) is a list with both fex IDs.
    - Usually something like C(1011).
    type: list
    elements: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

notes:
- The C(tenant), C(ap), C(epg) used must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant), M(cisco.aci.aci_ap), M(cisco.aci.aci_epg) modules can be used for this.
seealso:
- module: cisco.aci.aci_tenant
- module: cisco.aci.aci_ap
- module: cisco.aci.aci_epg
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(fv:RsPathAtt).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Bruno Calogero (@brunocalogero)
- Marcel Zehnder (@maercu)
"""

EXAMPLES = r"""
- name: Deploy Static Path binding for given EPG
  cisco.aci.aci_static_binding_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: accessport-code-cert
    ap: accessport_code_app
    epg: accessport_epg1
    encap_id: 222
    deploy_immediacy: lazy
    interface_mode: untagged
    interface_type: switch_port
    pod_id: 1
    leafs: 101
    interface: '1/7'
    state: present
  delegate_to: localhost

- name: Remove Static Path binding for given EPG
  cisco.aci.aci_static_binding_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: accessport-code-cert
    ap: accessport_code_app
    epg: accessport_epg1
    interface_type: switch_port
    pod: 1
    leafs: 101
    interface: '1/7'
    state: absent
  delegate_to: localhost

- name: Get specific Static Path binding for given EPG
  cisco.aci.aci_static_binding_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: accessport-code-cert
    ap: accessport_code_app
    epg: accessport_epg1
    interface_type: switch_port
    pod: 1
    leafs: 101
    interface: '1/7'
    state: query
  delegate_to: localhost
  register: query_result
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec

INTERFACE_MODE_MAPPING = {
    "802.1p": "native",
    "access": "untagged",
    "native": "native",
    "regular": "regular",
    "tagged": "regular",
    "trunk": "regular",
    "untagged": "untagged",
}

INTERFACE_TYPE_MAPPING = dict(
    fex="topology/pod-{pod_id}/paths-{leafs}/extpaths-{extpaths}/pathep-[eth{interface}]",
    fex_port_channel="topology/pod-{pod_id}/paths-{leafs}/extpaths-{extpaths}/pathep-[{interface}]",
    fex_vpc="topology/pod-{pod_id}/protpaths-{leafs}/extprotpaths-{extpaths}/pathep-[{interface}]",
    port_channel="topology/pod-{pod_id}/paths-{leafs}/pathep-[{interface}]",
    switch_port="topology/pod-{pod_id}/paths-{leafs}/pathep-[eth{interface}]",
    vpc="topology/pod-{pod_id}/protpaths-{leafs}/pathep-[{interface}]",
)

# TODO: change 'deploy_immediacy' to 'resolution_immediacy' (as seen in aci_epg_to_domain)?


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"]),  # Not required for querying all objects
        ap=dict(type="str", aliases=["app_profile", "app_profile_name"]),  # Not required for querying all objects
        epg=dict(type="str", aliases=["epg_name"]),  # Not required for querying all objects
        description=dict(type="str", aliases=["descr"]),
        encap_id=dict(type="int", aliases=["vlan", "vlan_id"]),
        primary_encap_id=dict(type="str", aliases=["primary_vlan", "primary_vlan_id"]),
        deploy_immediacy=dict(type="str", choices=["immediate", "lazy"]),
        interface_mode=dict(
            type="str", choices=["802.1p", "access", "native", "regular", "tagged", "trunk", "untagged"], aliases=["interface_mode_name", "mode"]
        ),
        interface_type=dict(type="str", default="switch_port", choices=["fex", "port_channel", "switch_port", "vpc", "fex_port_channel", "fex_vpc"]),
        pod_id=dict(type="int", aliases=["pod", "pod_number"]),  # Not required for querying all objects
        leafs=dict(type="list", elements="str", aliases=["leaves", "nodes", "paths", "switches"]),  # Not required for querying all objects
        interface=dict(type="str"),  # Not required for querying all objects
        extpaths=dict(type="list", elements="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["interface_type", "fex", ["extpaths"]],
            ["interface_type", "fex_vpc", ["extpaths"]],
            ["interface_type", "fex_port_channel", ["extpaths"]],
            ["state", "absent", ["ap", "epg", "interface", "leafs", "pod_id", "tenant"]],
            ["state", "present", ["ap", "encap_id", "epg", "interface", "leafs", "pod_id", "tenant"]],
        ],
    )

    tenant = module.params.get("tenant")
    ap = module.params.get("ap")
    epg = module.params.get("epg")
    description = module.params.get("description")
    encap_id = module.params.get("encap_id")
    primary_encap_id = module.params.get("primary_encap_id")
    deploy_immediacy = module.params.get("deploy_immediacy")
    interface_mode = module.params.get("interface_mode")
    interface_type = module.params.get("interface_type")
    pod_id = module.params.get("pod_id")
    leafs = module.params.get("leafs")
    interface = module.params.get("interface")
    extpaths = module.params.get("extpaths")
    state = module.params.get("state")

    aci = ACIModule(module)

    if leafs is not None:
        # Process leafs, and support dash-delimited leafs
        leafs = []
        for leaf in module.params.get("leafs"):
            # Users are likely to use integers for leaf IDs, which would raise an exception when using the join method
            leafs.extend(str(leaf).split("-"))
        if len(leafs) == 1:
            if interface_type in ["vpc", "fex_vpc"]:
                aci.fail_json(msg='A interface_type of "vpc" requires 2 leafs')
            leafs = leafs[0]
        elif len(leafs) == 2:
            if interface_type not in ["vpc", "fex_vpc"]:
                aci.fail_json(
                    msg='The interface_types "switch_port", "port_channel", and "fex" \
                    do not support using multiple leafs for a single binding'
                )
            leafs = "-".join(leafs)
        else:
            aci.fail_json(msg='The "leafs" parameter must not have more than 2 entries')

    if extpaths is not None:
        # Process extpaths, and support dash-delimited extpaths
        extpaths = []
        for extpath in module.params.get("extpaths"):
            # Users are likely to use integers for extpaths IDs, which would raise an exception when using the join method
            extpaths.extend(str(extpath).split("-"))
        if len(extpaths) == 1:
            if interface_type == "fex_vpc":
                aci.fail_json(msg='A interface_type of "fex_vpc" requires 2 extpaths')
            extpaths = extpaths[0]
        elif len(extpaths) == 2:
            if interface_type != "fex_vpc":
                aci.fail_json(
                    msg='The interface_types "fex" \
                    and "fex_port_channel" do not support using multiple extpaths for a single binding'
                )
            extpaths = "-".join(extpaths)
        else:
            aci.fail_json(msg='The "extpaths" parameter must not have more than 2 entries')

    if encap_id is not None:
        if encap_id not in range(1, 4097):
            aci.fail_json(msg="Valid VLAN assignments are from 1 to 4096")
        encap_id = "vlan-{0}".format(encap_id)

    if primary_encap_id is not None:
        try:
            primary_encap_id = int(primary_encap_id)
            if isinstance(primary_encap_id, int) and primary_encap_id in range(1, 4097):
                primary_encap_id = "vlan-{0}".format(primary_encap_id)
            else:
                aci.fail_json(msg="Valid VLAN assignments are from 1 to 4096 or unknown.")
        except Exception as e:
            if isinstance(primary_encap_id, str) and primary_encap_id != "unknown":
                aci.fail_json(msg="Valid VLAN assignments are from 1 to 4096 or unknown. %s" % e)

    static_path = INTERFACE_TYPE_MAPPING[interface_type].format(pod_id=pod_id, leafs=leafs, extpaths=extpaths, interface=interface)

    path_target_filter = {}
    if pod_id is not None and leafs is not None and interface is not None and (interface_type != "fex" or extpaths is not None):
        path_target_filter = {"tDn": static_path}

    if interface_mode is not None:
        interface_mode = INTERFACE_MODE_MAPPING[interface_mode]

    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="fvAp",
            aci_rn="ap-{0}".format(ap),
            module_object=ap,
            target_filter={"name": ap},
        ),
        subclass_2=dict(
            aci_class="fvAEPg",
            aci_rn="epg-{0}".format(epg),
            module_object=epg,
            target_filter={"name": epg},
        ),
        subclass_3=dict(
            aci_class="fvRsPathAtt",
            aci_rn="rspathAtt-[{0}]".format(static_path),
            module_object=static_path,
            target_filter=path_target_filter,
        ),
    )

    aci.get_existing()

    if state == "present":
        aci.payload(
            aci_class="fvRsPathAtt",
            class_config=dict(
                descr=description,
                encap=encap_id,
                primaryEncap=primary_encap_id,
                instrImedcy=deploy_immediacy,
                mode=interface_mode,
                tDn=static_path,
            ),
        )

        aci.get_diff(aci_class="fvRsPathAtt")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2024
# Hen Yaish <hyaish@redhat.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_imagesku_info

version_added: "2.4.0"

short_description: Get compute-related Image SKUs list

description:
    - Get details for compute-related resource Image SKUs.

options:
    location:
        description:
            - A region supported by current subscription.
        type: str
    offer:
        description:
            - The `Offer` refers to the specific product line of the operating system or software.
            - The offer usually includes multiple versions or configurations of the product.
        type: str
    publisher:
        description:
            - The `Publisher` is the entity that provides the OS image. It could be Microsoft, Canonical (for Ubuntu), Red Hat, etc.
            - It is a mandatory field that identifies the provider of the image.
        type: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Hen Yaish (@hyaish)

'''

EXAMPLES = '''
- name: Gather Resource Group info
  azure.azcollection.azure_rm_resourcegroup_info:
    name: "{{ resource_group }}"
  register: rg_info

- name: List available Image SKUs for 0001-com-ubuntu-server-focal
  azure.azcollection.azure_rm_imagesku_info:
    location: westus
    offer: 0001-com-ubuntu-server-focal
    publisher: Canonical
  register: available_image_skus_results
'''

RETURN = '''
available_skus:
    description:
        - List of Azure image SKUs for provisioning virtual machines.
    returned: always
    type: complex
    contains:
        name:
            description:
                - The specific SKU name or version.
            returned: always
            type: str
            sample: "20_04-lts"
        id:
            description:
                - The full Azure resource ID for the SKU.
            returned: always
            type: str
            sample: "0001-com-ubuntu-server-focal/Skus/20_04-lts"
        location:
            description:
                - The Azure region where the SKU is available.
            returned: always
            type: str
            sample: "westus"
        automatic_os_upgrade_supported:
            description:
                - Whether automatic OS upgrades are supported for this SKU.
            returned: always
            type: bool
            sample: false
        restrictions:
            description:
                - Restrictions that may apply to the use of this image SKU, such as region limitations or feature incompatibility.
            returned: always
            type: complex
            contains:
                type:
                    description:
                        - The type of restriction, which could include location-based restrictions or hardware compatibility issues.
                    type: str
                    returned: always
                    sample: "location"
                restriction_values:
                    description:
                        - A list of restricted locations, regions, or zones where the image SKU cannot be used.
                    type: list
                    returned: always
                    sample: ["eastus", "westeurope"]
                values:
                    description:
                        - A list of restricted locations, regions, or zones where the image SKU cannot be used.
                        - This return value has been deprecated and will be removed in a release after
                          2027-05-01. Use C(restriction_values) instead.
                    type: list
                    returned: always
                    sample: ["eastus", "westeurope"]
                reason_code:
                    description:
                        - The reason for the restriction, such as quota limitations or specific hardware requirements.
                    type: str
                    sample: "NotSupported"
                restriction_info:
                    description:
                        - Additional information about the restrictions, such as unsupported regions or features.
                    returned: always
                    type: complex
                    contains:
                        locations:
                            description:
                                - Locations where this SKU is restricted or unavailable.
                            type: list
                            sample: ["eastus"]
                        zones:
                            description:
                                - Availability zones within the region where this SKU is restricted.
                            type: list
                            sample: ["1", "2"]
'''


from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
try:
    from azure.mgmt.compute import ComputeManagementClient
    from azure.core.exceptions import HttpResponseError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMImageskuInfo(AzureRMModuleBase):
    def __init__(self):

        self.module_arg_spec = dict(
            location=dict(type='str'),
            publisher=dict(type='str'),
            offer=dict(type='str'),
        )

        self.results = dict(
            available_skus=[],
            count=0
        )
        self.location = None
        self.publisher = None
        self.offer = None
        super(AzureRMImageskuInfo, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                  supports_check_mode=True,
                                                  supports_tags=False)

    def exec_module(self, **kwargs):
        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        available_skus = self.list_skus()
        self.results['available_skus'] = available_skus
        self.results['count'] = len(available_skus)
        self.module.deprecate(
            "The 'values' return key in 'restrictions' is deprecated. Use 'restriction_values' instead.",
            date="2027-05-01",
            collection_name="azure.azcollection",
        )
        return self.results

    def list_skus(self):
        try:
            compute_client = self.get_mgmt_svc_client(ComputeManagementClient,
                                                      base_url=self._cloud_environment.endpoints.resource_manager,
                                                      api_version='2021-07-01')
            skus_result = compute_client.virtual_machine_images.list_skus(location=self.location,
                                                                          publisher_name=self.publisher,
                                                                          offer=self.offer)
            available_skus = []

            for sku_info in skus_result:
                sku_dict = sku_info.as_dict()
                for restriction in sku_dict.get('restrictions', []):
                    if 'values' in restriction:
                        restriction['restriction_values'] = restriction['values']
                available_skus.append(sku_dict)
            return available_skus

        except HttpResponseError as e:
            # Handle exceptions
            raise e


def _match_location(loc, locations):
    return next((x for x in locations if x.lower() == loc.lower()), None)


def _is_sku_available(sku_info, zone):
    """
    The SKU is unavailable in the following cases:
    1. regional restriction and the region is restricted
    2. parameter "zone" is input which indicates only showing skus with availability zones.
       Meanwhile, zonal restriction and all zones are restricted
    """
    is_available = True
    is_restrict_zone = False
    is_restrict_location = False
    if not sku_info.restrictions:
        return is_available
    for restriction in sku_info.restrictions:
        if restriction.reason_code == 'NotAvailableForSubscription':
            if restriction.type == 'Zone' and not (
                    set(sku_info.location_info[0].zones or []) - set(restriction.restriction_info.zones or [])):
                is_restrict_zone = True
            if restriction.type == 'Location' and (
                    sku_info.location_info[0].location in (restriction.restriction_info.locations or [])):
                is_restrict_location = True
            if is_restrict_location or (is_restrict_zone and zone):
                is_available = False
                break
    return is_available


def main():
    AzureRMImageskuInfo()


if __name__ == '__main__':
    main()
